# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides a command line interface for running Cicada programs.
"""

import argparse
import logging
import os
import signal
import subprocess
import sys
import tempfile
import urllib.parse

import netifaces
import numpy

import cicada
from cicada.additive import AdditiveProtocolSuite
from cicada.communicator import SocketCommunicator
from cicada.communicator.socket.connect import geturl
from cicada.logger import Logger


def get_environment(world_size, rank, address, root_address, identity, trusted):
    env = {
        "CICADA_WORLD_SIZE": str(world_size),
        "CICADA_RANK": str(rank),
        "CICADA_ADDRESS": address,
        "CICADA_ROOT_ADDRESS": root_address,
    }

    if identity:
        env["CICADA_IDENTITY"] = identity
    if trusted:
        env["CICADA_TRUSTED"] = ",".join(trusted)

    return env


def log_command(command, env, log):
    log.info(f"Command: {' '.join(command)}")
    log.info(f"  Environment:")
    for key, value in env.items():
        if key.startswith("CICADA_"):
            log.info(f"    {key}={value}")


def public_ip():
    """Return a public-facing IP address."""
    gateway, interface = netifaces.gateways()["default"][netifaces.AF_INET]
    return netifaces.ifaddresses(interface)[netifaces.AF_INET][0]["addr"]


def basic_frontend(arguments, players, log):
    processes = []
    for world_size, rank, address, root_address, identity, trusted in players:
        env = os.environ.copy()
        env.update(get_environment(world_size, rank, address, root_address, identity, trusted))

        command = [sys.executable]
        if arguments.inspect:
            command += ["-i"]
        command += [arguments.program]
        command += arguments.args

        log_command(command, env, log)

        if not arguments.dry_run:
            processes.append(subprocess.Popen(command, env=env))

    if not arguments.dry_run:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        for process in processes:
            process.wait()


def tmux_panes_frontend(arguments, players, log):
    command = []
    for world_size, rank, address, root_address, identity, trusted in players:
        env = get_environment(world_size, rank, address, root_address, identity, trusted)

        if rank == 0:
            command += ["tmux", "new-session"]
        else:
            command += [";", "split-window", "-v", "-d"]

        for name, value in env.items():
            command += ["-e", f"{name}={value}"]
        command += [sys.executable]
        if arguments.inspect:
            command += ["-i"]
        command += [arguments.program]
        command += arguments.args

    command += [";", "select-layout", arguments.tmux_layout]

    log.info(f"Command: {' '.join(command)}")

    processes = []
    if not arguments.dry_run:
        processes.append(subprocess.Popen(command))

    if not arguments.dry_run:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        for process in processes:
            process.wait()


def tmux_windows_frontend(arguments, players, log):
    command = []
    for world_size, rank, address, root_address, identity, trusted in players:
        env = get_environment(world_size, rank, address, root_address, identity, trusted)

        if rank == 0:
            command += ["tmux", "new-session"]
        else:
            command += [";", "new-window", "-d"]

        command += ["-n", f"rank-{rank}"]
        for name, value in env.items():
            command += ["-e", f"{name}={value}"]
        command += [sys.executable]
        if arguments.inspect:
            command += ["-i"]
        command += [arguments.program]
        command += arguments.args

    log.info(f"Command: {' '.join(command)}")

    processes = []
    if not arguments.dry_run:
        processes.append(subprocess.Popen(command))

    if not arguments.dry_run:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        for process in processes:
            process.wait()


def xterm_frontend(arguments, players, log):
    processes = []
    for world_size, rank, address, root_address, identity, trusted in players:
        env = os.environ.copy()
        env.update(get_environment(world_size, rank, address, root_address, identity, trusted))

        command = ["xterm", "-e"]
        command += [sys.executable]
        if arguments.inspect:
            command += ["-i"]
        command += [arguments.program]
        command += arguments.args

        log_command(command, env, log)

        if not arguments.dry_run:
            processes.append(subprocess.Popen(command, env=env))

    if not arguments.dry_run:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        for process in processes:
            process.wait()


frontends = {
    "basic": basic_frontend,
    "tmux": tmux_panes_frontend,
    "tmux-windows": tmux_windows_frontend,
    "xterm": xterm_frontend,
}


parser = argparse.ArgumentParser(description="Cicada MPC tools.")
subparsers = parser.add_subparsers(title="commands (choose one)", dest="command")

# certificate-info
certificate_info_subparser = subparsers.add_parser("certificate-info", help="Display information about a TLS certificate.")
certificate_info_subparser.add_argument("path", help="Path to the certificate file.")

# credentials
credentials_subparser = subparsers.add_parser("credentials", help="Generate player credentials for TLS encryption.")
credentials_subparser.add_argument("--certificate", default="player-{rank}.cert", help="Output certificate file. Default: %(default)s")
credentials_subparser.add_argument("--country", default="US", help="Certificate country. Default: %(default)s")
credentials_subparser.add_argument("--days", type=int, default=365, help="Length of time the certificate will be valid. Default: %(default)s")
credentials_subparser.add_argument("--email", default=None, help="Certificate email. Default: %(default)s")
credentials_subparser.add_argument("--identity", default="player-{rank}.pem", help="Output identity (private key and certificate) file. Default: %(default)s")
credentials_subparser.add_argument("--locality", default="Albuquerque", help="Certificate locality. Default: %(default)s")
credentials_subparser.add_argument("--name", default=None, help="Common name. Default: based on player rank.")
credentials_subparser.add_argument("--organization", default="Cicada", help="Certificate organization. Default: %(default)s")
credentials_subparser.add_argument("--rank", required=True, help="Player rank.")
credentials_subparser.add_argument("--state", default="New Mexico", help="Certificate state. Default: %(default)s")
credentials_subparser.add_argument("--unit", default=None, help="Certificate organizational unit. Default: %(default)s")

# generate-shares
generate_subparser = subparsers.add_parser("generate-shares", help="Generate secret shares for pedagogy.")
generate_subparser.add_argument("--order", "-m", type=int, default=18446744073709551557, help="Field modulus. Default: %(default)s")
generate_subparser.add_argument("--precision", "-p", type=int, default=16, help="Fractional precision. Default: %(default)s")
generate_subparser.add_argument("--world-size", "-n", type=int, default=3, help="Number of players. Default: %(default)s")
generate_subparser.add_argument("secret", help="Secret value.")

# run
run_subparser = subparsers.add_parser("run", help="Run all Cicada processes on the local machine.")
run_subparser.add_argument("--dry-run", "-y", action="store_true", help="Don't start actual processes.")
run_subparser.add_argument("--frontend", "-f", choices=frontends.keys(), default="basic", help="Frontend to execute processes.")
run_subparser.add_argument("--identity", default="player-{rank}.pem", help="Player private key and certificate in PEM format. Default: %(default)s file, if it exists.")
run_subparser.add_argument("--inspect", "-i", action="store_true", help="Start a Python prompt after running program.")
run_subparser.add_argument("--trusted", default="player-{rank}.cert", help="Trusted certificates in PEM format. Default: %(default)s files, if they exist.")
run_subparser.add_argument("--root-address", default="tcp://127.0.0.1:25252", help="Root address.  Default: %(default)s")
run_subparser.add_argument("--tmux-layout", default="even-vertical", choices=["even-horizontal", "even-vertical", "tiled"], help="Pane layout for the tmux frontend. Default: %(default)s")
run_subparser.add_argument("--world-size", "-n", type=int, default=3, help="Number of players. Default: %(default)s")
run_subparser.add_argument("program", help="Program to execute.")
run_subparser.add_argument("args", nargs=argparse.REMAINDER, help="Program arguments.")

# start
start_subparser = subparsers.add_parser("start", help="Start one Cicada process.")
start_subparser.add_argument("--dry-run", "-y", action="store_true", help="Don't start actual processes.")
start_subparser.add_argument("--address", default=None, help=f"Network address. Default: tcp://{public_ip()}:25252 for player 0, otherwise tcp://{public_ip()}")
start_subparser.add_argument("--identity", default="player-{rank}.pem", help="Player private key and certificate in PEM format. Default: %(default)s file, if it exists.")
start_subparser.add_argument("--inspect", "-i", action="store_true", help="Start a Python prompt after running program.")
start_subparser.add_argument("--trusted", default="player-{rank}.cert", help="Trusted certificates in PEM format. Default: %(default)s files, if they exist.")
start_subparser.add_argument("--root-address", default=None, help="Root address. Default: same as --address for player 0, required otherwise.")
start_subparser.add_argument("--rank", type=int, required=True, help="Player rank.")
start_subparser.add_argument("--world-size", "-n", type=int, default=3, help="Number of players. Default: %(default)s")
start_subparser.add_argument("program", help="Program to execute.")
start_subparser.add_argument("args", nargs=argparse.REMAINDER, help="Program arguments.")

# version
version_subparser = subparsers.add_parser("version", help="Print the Cicada version.")


def main():
    arguments = parser.parse_args()

    if arguments.command is None:
        parser.print_help()

    logging.basicConfig(level=logging.INFO)
    log = logging.getLogger()
    log.name = os.path.basename(sys.argv[0])


    # certificate-info
    if arguments.command == "certificate-info":
        subprocess.run(["openssl", "x509", "-in", arguments.path, "-text"], check=True)

    # credentials
    if arguments.command == "credentials":
        rank = arguments.rank

        if arguments.name is None:
            arguments.name = f"Player {rank}"

        subj = ""
        subj += f"/C={arguments.country}" if arguments.country else ""
        subj += f"/ST={arguments.state}" if arguments.state else ""
        subj += f"/L={arguments.locality}" if arguments.locality else ""
        subj += f"/O={arguments.organization}" if arguments.organization else ""
        subj += f"/OU={arguments.unit}" if arguments.unit else ""
        subj += f"/emailAddress={arguments.email}" if arguments.email else ""
        subj += f"/CN={arguments.name}"

        certificate_path = arguments.certificate.format(rank=rank)
        identity_path = arguments.identity.format(rank=rank)

        fd, key_path = tempfile.mkstemp()
        os.close(fd)

        subprocess.check_call(["openssl", "genrsa", "-out", key_path, "2048"])
        subprocess.check_call(["openssl", "req", "-new", "-key", key_path, "-x509", "-subj", subj, "-out", certificate_path, "-days", str(arguments.days)])
        with open(identity_path, "wb") as target:
            with open(key_path, "rb") as source:
                target.write(source.read())
            with open(certificate_path, "rb") as source:
                target.write(source.read())
        os.remove(key_path)

    # generate-shares
    if arguments.command == "generate-shares":
        logging.basicConfig(level=logging.INFO)

        def main(communicator, precision, order, secret):
            log = Logger(logging.getLogger(), communicator)

            protocol = AdditiveProtocolSuite(communicator, order=order, encoding=cicada.encoding.FixedPoint(precision=precision))
            secret = numpy.array(secret, dtype=float)
            share = protocol.share(src=0, secret=secret, shape=secret.shape)

            log.info(f"Player {communicator.rank} secret share: {share.storage}")

        SocketCommunicator.run(world_size=arguments.world_size, fn=main, args=(arguments.precision, arguments.order, arguments.secret))

    # run
    if arguments.command == "run":
        world_size = arguments.world_size
        if world_size < 1:
            run_subparser.error("--world-size must be greater than zero.")

        root_address = urllib.parse.urlparse(arguments.root_address)
        if root_address.scheme not in ["file", "tcp"]:
            run_subparser.error(f"--root-address scheme must be file or tcp, got {root_address.scheme} instead.")
        if root_address.scheme == "tcp" and root_address.port is None:
            run_subparser.error("--root-address must specify a port number when the scheme is tcp.")

        addresses = []
        for rank in range(world_size):
            if rank == 0:
                addresses.append(arguments.root_address)
            else:
                if root_address.scheme == "file":
                    fd, path = tempfile.mkstemp()
                    os.close(fd)
                    addresses.append(f"file://{path}")
                if root_address.scheme == "tcp":
                    addresses.append(f"tcp://{root_address.hostname}")

        players = []
        for rank, address in enumerate(addresses):
            identity = arguments.identity.format(rank=rank, world_size=world_size)
            if not os.path.exists(identity):
                identity = ""

            trusted = []
            for index in range(world_size):
                trust = arguments.trusted.format(rank=index, world_size=world_size)
                if index != rank and os.path.exists(trust):
                    trusted.append(trust)

            players.append((world_size, rank, address, addresses[0], identity, trusted))

        frontend = frontends[arguments.frontend]
        frontend(arguments, players, log)

    # start
    if arguments.command == "start":
        world_size = arguments.world_size
        if world_size < 1:
            start_subparser.error("--world-size must be greater than zero.")

        rank = arguments.rank
        if rank < 0 or rank >= world_size:
            start_subparser.error(f"--rank must be in the range [0, {world_size}).")

        address = arguments.address
        if address is None:
            address = f"tcp://{public_ip()}:25252" if rank == 0 else f"tcp://{public_ip()}"
        address = urllib.parse.urlparse(address)
        address = address.geturl()

        root_address = arguments.root_address
        if root_address is None:
            if rank == 0:
                root_address = address
            else:
                start_subparser.error(f"--root-address must be specified.")
        root_address = urllib.parse.urlparse(root_address)
        root_address = root_address.geturl()

        identity = arguments.identity.format(rank=rank, world_size=world_size)
        if not os.path.exists(identity):
            identity = ""

        trusted = []
        for index in range(world_size):
            trust = arguments.trusted.format(rank=index, world_size=world_size)
            if index != rank and os.path.exists(trust):
                trusted.append(trust)

        players = [(world_size, rank, address, root_address, identity, trusted)]
        frontend = frontends["basic"]
        frontend(arguments, players, log)

    # version
    if arguments.command == "version":
        print(cicada.__version__)





# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import pickle

from cicada.communicator import SocketCommunicator
from cicada.communicator.socket import connect


logging.basicConfig(level=logging.INFO)
logging.getLogger("cicada.communicator").setLevel(logging.INFO)

world_size = int(os.environ.get("CICADA_WORLD_SIZE"))
rank = int(os.environ.get("CICADA_RANK"))
address = os.environ.get("CICADA_ADDRESS")
root_address = os.environ.get("CICADA_ROOT_ADDRESS")

timer = connect.Timer(threshold=300)
listen_socket = connect.listen(address=address, rank=rank, timer=timer)
sockets = connect.rendezvous(listen_socket=listen_socket, root_address=root_address, world_size=world_size, rank=rank, timer=timer)
with SocketCommunicator(sockets=sockets) as communicator:
    print(f"Player {communicator.rank} waiting for commands.")

    listen_socket.setblocking(True)
    while True:
        client, addr = listen_socket.accept()
        command = pickle.loads(client.recv(4096))
        print(addr, command)





# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for communicating using the builtin :mod:`socket` module.
"""

import json
import logging
import numbers
import os
import pickle
import pprint
import select
import socket
import ssl
import time
import urllib.parse

import pynetstring


class CommunicatorEvents(object):
    def __init__(self, name, rank):
        self.name = name
        self.rank = rank

    def filter(self, record):
        record.comm = self.name
        record.rank = self.rank
        return True


class EncryptionFailed(Exception):
    """Raised if an encrypted connection can't be established with another player."""
    pass


class LoggerAdapter(logging.LoggerAdapter):
    """Wraps a Python logger for consistent formatting of communicator log entries.

    logger: :class:`logging.Logger`, required
        Python logger to wrap.
    name: :class:`str`, required
        Communicator name.
    rank: :class:`int`, required
        Communicator rank
    """
    def __init__(self, logger, name, rank):
        super().__init__(logger, extra={"name": name, "rank": rank})

    def process(self, msg, kwargs):
        return message(self.extra["name"], self.extra["rank"], msg), kwargs


class NetstringSocket(object):
    """Message-oriented socket that uses the Netstrings protocol."""
    def __init__(self, sock):
        self._socket = sock
        self._decoder = pynetstring.Decoder()
        self._messages = []
        self._sent_bytes = 0
        self._sent_messages = 0
        self._received_bytes = 0
        self._received_messages = 0

    def close(self):
        """Close the underlying socket."""
        self._socket.close()

    def feed(self):
        """Read data from the underlying socket, decoding whatever is available."""
        raw = self._socket.recv(4096)
        messages = self._decoder.feed(raw)
        self._received_bytes += len(raw)
        self._received_messages += len(messages)
        self._messages += messages

    @property
    def family(self):
        """Return the underlying socket's address family.

        See :attr:`socket.socket.family`.
        """
        return self._socket.family

    def fileno(self):
        """Return the file descriptor for the underlying socket.

        This allows :class:`NetstringSocket` to be used with :func:`select.select`.
        """
        return self._socket.fileno()

    def getsockname(self):
        """Return the underlying socket's address.

        See :meth:`socket.socket.getsockname`.
        """
        return self._socket.getsockname()

    def messages(self):
        """Return every message that has been received, if any."""
        result = self._messages
        self._messages = []
        return result

    def next_message(self, *, timeout):
        """Return the next available message, if any."""
        if not self._messages:
            ready, _, _ = select.select([self], [], [], timeout)
            if ready:
                self.feed()
        return self._messages.pop(0) if self._messages else None

    def send(self, msg):
        """Send a message."""
        raw = pynetstring.encode(msg)
        self._sent_bytes += len(raw)
        self._sent_messages += 1
        while raw:
            _, ready, _ = select.select([], [self], [])
            if ready:
                sent = self._socket.send(raw)
                raw = raw[sent:]

    @property
    def stats(self):
        """Return a :class:`dict` containing statistics.

        Returns the number of bytes and messages that have been sent and received.
        """
        return {
            "sent": {"bytes": self._sent_bytes, "messages": self._sent_messages},
            "received": {"bytes": self._received_bytes, "messages": self._received_messages},
            }


class Timeout(Exception):
    """Raised when an operation has timed-out."""
    pass


class Timer(object):
    """Tracks elapsed time.

    Parameters
    ----------
    threshold: :class:`numbers.Number`, required
        The maximum of number of elapsed seconds before this timer will be expired.
    """
    def __init__(self, threshold):
        self._start = time.time()
        self._threshold = threshold

    @property
    def elapsed(self):
        """Return the number of seconds since the :class:`Timer` was created."""
        return time.time() - self._start


    @property
    def expired(self):
        """Returns :any:`True` if the elapsed time has exceeded a threshold."""
        if self._threshold and (self.elapsed > self._threshold):
            return True
        return False


class TokenMismatch(Exception):
    """Raised when players can't agree on a token for communicator creation."""
    pass


def certformat(cert, indent=""):
    """Format an X509 certificate for easy reading.

    Parameters
    ----------
    cert: :class:`dict`, required
        X509 certificate returned from :meth:`ssl.SSLSocket.getpeercert` or similar.
    indent: :class:`str`, optional
        Optional indenting to be added to the output.

    Returns
    -------
    prettycert: :class:`str`
        The contents of the certificate, formatted for easy reading by humans.
    """
    mapping = {"countryName": "C", "stateOrProvinceName": "ST", "localityName": "L", "organizationName": "O", "commonName": "CN"}

    issuer = [b for a in cert.get("issuer") for b in a]
    issuer = [f"{mapping.get(item[0], item[0])}={item[1]}" for item in issuer]
    issuer = ", ".join(issuer)

    subject = [b for a in cert.get("subject") for b in a]
    subject = [f"{mapping.get(item[0], item[0])}={item[1]}" for item in subject]
    subject = ", ".join(subject)

    return f"""{indent}Serial: {cert.get("serialNumber")}
{indent}Issuer: {issuer}
{indent}Subject: {subject}
{indent}Valid From: {cert.get("notBefore")} Until: {cert.get("notAfter")}"""


def connect(*, address, rank, other_rank, name="world", tls=None):
    """Given an address, create a socket and make a connection.

    Parameters
    ----------
    address: :class:`urllib.parse.ParseResult`, required
        The address URL.
    rank: :class:`int`, required
        Rank of the calling player.
    other_rank: :class:`int`, required
        Rank of the other player (the one we're connecting to).
    name: :class:`str`, optional
        Human readable label used for logging and error messages. Typically,
        this should be the same name that will be eventually used by a
        communicator instance. Defaults to "world".
    tls: pair of :class:`ssl.SSLContext`, optional
        If provided, the returned sockets will implement transport layer
        security.  Callers must provide a sequence containing one context
        configured for server connections, and one for configured for client
        connections, in that order.

    Returns
    -------
    socket: :class:`NetstringSocket`
        The newly-connected socket, ready for use.
    """
    log = getLogger(__name__, name, rank)

    if address.scheme == "file":
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(address.path)

    elif address.scheme == "tcp":
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((address.hostname, address.port))

    else:
        raise ValueError(f"address.scheme must be file or tcp, got {address.scheme} instead.") # pragma: no cover

    sock.setblocking(True)


    # Optionally setup TLS.
    if tls is not None:
        sock = tls[1].wrap_socket(sock, server_side=False)
        log.info(f"{geturl(sock)} connected to player {other_rank} {getpeerurl(sock)}, received certificate:\n{certformat(sock.getpeercert(), indent='    ')}")
    else:
        log.info(f"{geturl(sock)} connected to player {other_rank} {getpeerurl(sock)}")

    return NetstringSocket(sock)


def direct(*, listen_socket, addresses, rank, name="world", timer=None, tls=None):
    """Create socket connections given a list of player addresses.

    Parameters
    ----------
    listen_socket: :class:`socket.socket`, required
        A nonblocking Python socket or compatible object, already bound and
        listening for connections.  Typically created with :func:`listen`.
    addresses: :class:`list` of :class:`str`, required
        List of address URLs for every player in rank order.
    rank: :class:`int`, required
        Rank of the calling player.
    name: :class:`str`, optional
        Human readable label used for logging and error messages. Typically,
        this should be the same name assigned to the communicator that will use
        the :func:`direct` outputs.  Defaults to "world".
    timer: :class:`Timer`, optional
        Determines the maximum time to wait for player connections.  Defaults
        to five seconds.
    tls: pair of :class:`ssl.SSLContext`, optional
        If provided, the returned sockets will implement transport layer
        security.  Callers must provide a sequence containing one context
        configured for server connections, and one for configured for client
        connections, in that order.

    Raises
    ------
    :class:`EncryptionFailed`
        If there are problems establishing an encrypted connection with another
        player.
    :class:`Timeout`
        If `timer` expires before all connections are established.
    :class:`ValueError`
        If there are problems with input parameters.

    Returns
    -------
    sockets: :class:`dict` of :class:`NetstringSocket`
        Dictionary mapping player ranks to connected sockets.
    """

    for address in addresses:
        if not isinstance(address, str):
            raise ValueError(message(name, rank, f"address must be a string, got {address} instead.")) # pragma: no cover

    addresses = [urllib.parse.urlparse(address) for address in addresses]

    for address in addresses:
        if address.scheme not in ["file", "tcp"]:
            raise ValueError(message(name, rank, f"address scheme must be file or tcp, got {address.scheme} instead.")) # pragma: no cover
        if address.scheme == "tcp" and not address.port:
            raise ValueError(message(name, rank, f"port must be specified for tcp addresses, got {address} instead.")) # pragma: no cover
        if address.scheme != addresses[0].scheme:
            raise ValueError(message(name, rank, f"address schemes must match.")) # pragma: no cover

    world_size = len(addresses)

    if not isinstance(rank, int):
        raise ValueError(message(name, rank, f"rank must be an integer, got {rank} instead.")) # pragma: no cover
    if rank < 0 or rank >= world_size:
        raise ValueError(message(name, rank, f"rank must be in the range [0, {world_size}), got {rank} instead.")) # pragma: no cover

    if not isinstance(name, str):
        raise ValueError(message(name, rank, f"name must be a string, got {name} instead.")) # pragma: no cover

    if timer is None:
        timer = Timer(threshold=5)
    if not isinstance(timer, Timer):
        raise ValueError(message(name, rank, f"timer must be an instance of Timer, got {timer} instead.")) # pragma: no cover

    # Setup logging
    log = getLogger(__name__, name, rank)
    log.info(f"direct connect with {[address.geturl() for address in addresses]}.")

    # Set aside storage for connections to the other players.
    players = {}

    # Players setup connections with each other, in rank order.
    for listener in range(0, world_size-1):
        if rank == listener:
            # Accept connections from higher-rank players.
            other_players = []
            while not timer.expired:
                if len(other_players) == world_size - rank - 1: # We don't connect to ourself.
                    break

                try:
                    ready = select.select([listen_socket], [], [], 0.1)
                    if ready:
                        sock, _ = listen_socket.accept()
                        if tls is not None:
                            sock = tls[0].wrap_socket(sock, server_side=True)
                            log.info(f"{geturl(sock)} accepted connection from {getpeerurl(sock)}, received certificate:\n{certformat(sock.getpeercert(), indent='    ')}")
                        else:
                            log.info(f"{geturl(sock)} accepted connection from {getpeerurl(sock)}")
                        sock = NetstringSocket(sock)
                        other_players.append(sock)
                except ssl.SSLError as e:
                    # There was a problem setting up an encrypted connection with
                    # the other player, so there's no point in continuing.
                    raise EncryptionFailed(message(name, rank, f"remote player failed to connect: {e}"))
                except Exception as e: # pragma: no cover
                    log.warning(f"exception listening for other players: {e}")
                    time.sleep(0.1)

            # Collect ranks from the other players.
            for other_player in other_players:
                while not timer.expired:
                    try:
                        raw_message = other_player.next_message(timeout=0.1)
                        if raw_message is not None:
                            other_rank = pickle.loads(raw_message)
                            players[other_rank] = other_player
                            log.debug(f"received rank from player {other_rank}.")
                            other_player.send(pickle.dumps("OK"))
                            break
                    except Exception as e: # pragma: no cover
                        log.warning(f"exception receiving player rank: {e}")
                        time.sleep(0.1)
                else: # pragma: no cover
                    raise Timeout(message(name, rank, "timeout waiting for player rank."))

        elif rank > listener:
            # Make a connection to the listener.
            while not timer.expired:
                try:
                    players[listener] = connect(address=addresses[listener], rank=rank, other_rank=listener, name=name, tls=tls)
                    break
                except ssl.SSLCertVerificationError as e:
                    # If this happens it means we couldn't verify the other
                    # player's certificate, so there's no point in continuing.
                    raise EncryptionFailed(message(name, rank, f"received invalid certificate from player {listener}."))
                except Exception as e: # pragma: no cover
                    log.warning(f"exception connecting to player {listener}: {e}")
                    time.sleep(0.5)
            else: # pragma: no cover
                raise Timeout(message(name, rank, f"timeout connecting to player {listener}."))

            # Send our rank to the listener.
            while not timer.expired:
                try:
                    players[listener].send(pickle.dumps(rank))
                    break
                except Exception as e: # pragma: no cover
                    log.warning(f"exception sending rank to player {listener}: {e}")
                    time.sleep(0.5)
            else: # pragma: no cover
                raise Timeout(message(name, rank, f"timeout sending rank to player {listener}."))

            # Receive a response from the listener.
            while not timer.expired:
                try:
                    raw_message = players[listener].next_message(timeout=0.1)
                    if raw_message is not None:
                        response = pickle.loads(raw_message)
                        log.debug(f"received response from player {listener}.")
                        break
                except Exception as e: # pragma: no cover
                    log.warning(f"exception waiting for response from player {listener}: {e}")
                    time.sleep(0.1)
            else: # pragma: no cover
                raise Timeout(message(name, rank, f"timeout waiting for response from player {listener}."))

    return players


def getLogger(name, comm, rank):
    logger = logging.getLogger(name)
    logger.addFilter(CommunicatorEvents(comm, rank))
    return LoggerAdapter(logger, comm, rank)


def getpeerurl(sock):
    """Return a socket's peer address as a URL.

    Parameters
    ----------
    sock: :class:`socket.socket`, required

    Returns
    -------
    url: :class:`str`
        The socket's local address as a URL.  For example:
        `"tcp://127.0.0.1:59678"` for TCP sockets, or `"file:///path/to/foo"` for
        Unix domain sockets.
    """
    if sock.family == socket.AF_UNIX:
        path = sock.getpeername()
        return f"file://{path}"

    if sock.family == socket.AF_INET:
        host, port = sock.getpeername()
        return f"tcp://{host}:{port}"

    raise ValueError(f"Unknown address family: {sock.family}") # pragma: no cover


def gettls(*, identity=None, trusted=None):
    """Construct a pair of :class:`ssl.SSLContext` instances.

    Parameters
    ----------
    identity: :class:`str`, optional
        Path to a private key and certificate in PEM format that will
        identify the local player.
    trusted: sequence of :class:`str`, optional
        Path to certificates in PEM format that will identify the other
        players.

    Returns
    -------
    tls: (server, client) tuple or :any:`None`
    """

    if identity and trusted:
        server = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        server.load_cert_chain(certfile=identity)
        for trust in trusted:
            server.load_verify_locations(trust)
        server.check_hostname=False
        server.verify_mode = ssl.CERT_REQUIRED

        client = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        client.load_cert_chain(certfile=identity)
        for trust in trusted:
            client.load_verify_locations(trust)
        client.check_hostname = False
        client.verify_mode = ssl.CERT_REQUIRED

        return (server, client)

    return None



def geturl(sock):
    """Return a socket's local address as a URL.

    Parameters
    ----------
    sock: :class:`socket.socket`, required

    Returns
    -------
    url: :class:`str`
        The socket's local address as a URL.  For example:
        `"tcp://127.0.0.1:59678"` for TCP sockets, or `"file:///path/to/foo"` for
        Unix domain sockets.
    """
    if sock.family == socket.AF_UNIX:
        path = sock.getsockname()
        return f"file://{path}"

    if sock.family == socket.AF_INET:
        host, port = sock.getsockname()
        return f"tcp://{host}:{port}"

    raise ValueError(f"Unknown address family: {sock.family}") # pragma: no cover


def listen(*, address, rank, name="world", timer=None):
    """Create a listening socket from a URL.

    Typically, callers should use this function to create a listening socket
    for use with either :func:`direct` or :func:`rendezvous`.

    Parameters
    ----------
    address: :class:`str`, required
        Address to use for listening, in the form of a URL.  For example:
        `"tcp://127.0.0.1:59478"` to create a TCP socket, or
        `"file:///path/to/foo"` to create a Unix domain socket.
    rank: :class:`int`, required
        Integer rank of the caller.  Used for logging and error messages.
    name: :class:`str`, optional
        Human readable label.  Used for logging and error messages.  Typically,
        this should match the name used elsewhere to create a communicator.
    timer: :class:`Timer`, optional
        Determines the maximum time to wait for socket creation.  Defaults to
        five seconds.

    Raises
    ------
    :class:`ValueError`
        If there are problems with input parameters.
    :class:`Timeout`
        If `timer` expires before all connections are established.

    Returns
    -------
    socket: :class:`socket.socket`
        A non-blocking socket, bound to `address`, listening for connections.
    """
    if not isinstance(address, str):
        raise ValueError(message(name, rank, f"address must be a string, got {address} instead.")) # pragma: no cover
    address = urllib.parse.urlparse(address)
    if address.scheme not in ["file", "tcp"]:
        raise ValueError(message(name, rank, f"address scheme must be file or tcp, got {address.scheme} instead.")) # pragma: no cover

    if not isinstance(rank, int):
        raise ValueError(message(name, rank, f"rank must be an integer, got {rank} instead.")) # pragma: no cover

    if not isinstance(name, str):
        raise ValueError(message(name, rank, f"name must be a string, got {name} instead.")) # pragma: no cover

    if timer is None:
        timer = Timer(threshold=5)
    if not isinstance(timer, Timer):
        raise ValueError(message(name, rank, f"timer must be an instance of Timer, got {timer} instead.")) # pragma: no cover

    # Setup logging
    log = getLogger(__name__, name, rank)

    # Create the socket.
    while not timer.expired:
        try:
            if address.scheme == "tcp":
                listen_socket = socket.create_server((address.hostname, address.port or 0))
                address = urllib.parse.urlparse(geturl(listen_socket)) # Recreate the address in case the port was assigned at random.
            elif address.scheme == "file":
                if os.path.exists(address.path):
                    os.unlink(address.path) # pragma: no cover
                listen_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                listen_socket.bind(address.path)
            listen_socket.setblocking(False)
            listen_socket.listen()
            log.info(f"listening to {address.geturl()} for connections.")
            break
        except Exception as e: # pragma: no cover
            log.warning(f"exception creating listening socket: {e}")
            time.sleep(0.1)
    else: # pragma: no cover
        raise Timeout(message(name, rank, f"timeout creating listening socket."))

    return listen_socket


def message(name, rank, msg):
    """Format a message for logging and error handling."""
    return f"Comm {name} player {rank} {msg}"


def rendezvous(*, listen_socket, root_address, world_size, rank, name="world", token=0, timer=None, tls=None):
    """Create socket connections given just the address of the root player.

    Parameters
    ----------
    listen_socket: :class:`socket.socket`, required
        A nonblocking Python socket or compatible object, already bound and
        listening for connections.  Typically created with :func:`listen`.
    root_address: :class:`str`, required
        URL address of the root (rank 0) player.  The address must be reachable
        by all of the other players.
    world_size: :class:`int`, required
        The number of players that will be members of this communicator.
    rank: :class:`int`, required
        The rank of the caller, in the range [0, world_size).
    name: :class:`str`, optional
        Human readable label used for logging and debugging. Typically, this
        should be the same name assigned to the communicator that will use
        the :func:`rendezvous` outputs.  Defaults to "world".
    token: :class:`object`, optional
        All players provide an arbitrary token at startup; every player token
        must match, or a :class:`TokenMismatch` exception will be raised.
    timer: :class:`Timer`, optional
        Determines the maximum time to wait for socket creation.  Defaults to
        five seconds.
    tls: pair of :class:`ssl.SSLContext`, optional
        If provided, the returned sockets will implement transport layer
        security.  Callers must provide a sequence containing one context
        configured for server connections, and one for configured for client
        connections, in that order.

    Raises
    ------
    :class:`EncryptionFailed`
        If there are problems establishing an encrypted connection with another
        player.
    :class:`Timeout`
        If `timer` expires before all connections are established.
    :class:`TokenMismatch`
        If every player doesn't call this function with the same token.
    :class:`ValueError`
        If there are problems with input parameters.

    Returns
    -------
    sockets: :class:`dict` of :class:`NetstringSocket`
        Dictionary mapping player ranks to connected sockets.
    """
    if not isinstance(root_address, str):
        raise ValueError(message(name, rank, f"root_address must be a string, got {root_address} instead.")) # pragma: no cover
    root_address = urllib.parse.urlparse(root_address)
    if root_address.scheme not in ["file", "tcp"]:
        raise ValueError(message(name, rank, f"root_address scheme must be file or tcp, got {root_address.scheme} instead.")) # pragma: no cover
    if rank == 0 and root_address.geturl() != geturl(listen_socket):
        raise ValueError(message(name, rank, f"Player 0 root_address {root_address} must match listen_socket.")) # pragma: no cover

    if not isinstance(world_size, int):
        raise ValueError(message(name, rank, f"world_size must be an integer, got {world_size} instead.")) # pragma: no cover
    if not world_size > 0:
        raise ValueError(message(name, rank, f"world_size must be greater than zero, got {world_size} instead.")) # pragma: no cover

    if not isinstance(rank, int):
        raise ValueError(message(name, rank, f"rank must be an integer, got {rank} instead.")) # pragma: no cover
    if rank < 0 or rank >= world_size:
        raise ValueError(message(name, rank, f"rank must be in the range [0, {world_size}), got {rank} instead.")) # pragma: no cover

    if not isinstance(name, str):
        raise ValueError(message(name, rank, f"name must be a string, got {name} instead.")) # pragma: no cover

    if timer is None:
        timer = Timer(threshold=5)
    if not isinstance(timer, Timer):
        raise ValueError(message(name, rank, f"timer must be an instance of Timer, got {timer} instead.")) # pragma: no cover

    # Setup logging
    log = getLogger(__name__, name, rank)
    log.info(f"rendezvous with {root_address.geturl()} from {geturl(listen_socket)}")

    # Set aside storage for connections to the other players.
    players = {}

    ###########################################################################
    # Phase 1: Every player (except root) makes a connection to root.

    if rank != 0:
        while not timer.expired:
            try:
                players[0] = connect(address=root_address, rank=rank, other_rank=0, name=name, tls=tls)
                break
            except ConnectionRefusedError as e: # pragma: no cover
                # This happens regularly, particularly when starting on
                # separate hosts, so no need to log a warning.
                time.sleep(0.1)
            except ssl.SSLCertVerificationError as e:
                # If this happens it means we couldn't verify the other
                # player's certificate, so there's no point in continuing.
                raise EncryptionFailed(message(name, rank, "received invalid certificate from player 0."))
            except Exception as e: # pragma: no cover
                log.warning(f"exception connecting to player 0: {e}")
                time.sleep(0.1)
        else: # pragma: no cover
            raise Timeout(message(name, rank, "timeout connecting to player 0."))

    ###########################################################################
    # Phase 2: Every player sends their listening address to root.

    if rank != 0:
        while not timer.expired:
            try:
                players[0].send(pickle.dumps((rank, geturl(listen_socket), token)))
                break
            except Exception as e: # pragma: no cover
                log.warning(f"exception sending address to player 0: {e}")
                time.sleep(0.1)
        else: # pragma: no cover
            raise Timeout(message(name, rank, f"timeout sending address to player 0."))

    ###########################################################################
    # Phase 3: Root gathers addresses from every player.

    if rank == 0:
        # Accept a connection from every player.
        other_players = []
        while not timer.expired:
            if len(other_players) == world_size - 1: # We don't connect to ourself.
                break

            try:
                ready = select.select([listen_socket], [], [], 0.1)
                if ready:
                    sock, _ = listen_socket.accept()
                    if tls is not None:
                        sock = tls[0].wrap_socket(sock, server_side=True)
                        log.info(f"{geturl(sock)} accepted connection from {getpeerurl(sock)}, received certificate:\n{certformat(sock.getpeercert(), indent='    ')}")
                    else:
                        log.debug(f"{geturl(sock)} accepted connection from {getpeerurl(sock)}.")
                    sock = NetstringSocket(sock)
                    other_players.append(sock)
            except ssl.SSLError as e:
                # There was a problem setting up an encrypted connection with
                # the other player, so there's no point in continuing.
                raise EncryptionFailed(message(name, rank, f"remote player failed to connect: {e}"))
            except BlockingIOError as e: # pragma: no cover
                # This happens regularly, particularly when starting on
                # separate hosts, so no need to log a warning.
                time.sleep(0.1)
            except Exception as e: # pragma: no cover
                log.warning(f"exception waiting for connections from other players: {type(e)} {e}")
                time.sleep(0.1)
        else: # pragma: no cover
            raise Timeout(message(name, rank, "timeout waiting for player connections."))

        # Collect an address from every player.
        addresses = {rank: (geturl(listen_socket), token)}
        for other_player in other_players:
            while not timer.expired:
                try:
                    raw_message = other_player.next_message(timeout=0.1)
                    if raw_message is not None:
                        other_rank, other_addr, other_token = pickle.loads(raw_message)
                        players[other_rank] = other_player
                        addresses[other_rank] = (other_addr, other_token)
                        log.debug(f"received address from player {other_rank}.")
                        break
                except Exception as e: # pragma: no cover
                    log.warning(f"exception receiving player address: {e}")
                    time.sleep(0.1)
            else: # pragma: no cover
                raise Timeout(message(name, rank, "timeout waiting for player address."))

    ###########################################################################
    # Phase 4: Root broadcasts the set of all addresses to every player.

    if rank == 0:
        for player in players.values():
            player.send(pickle.dumps(addresses))

    ###########################################################################
    # Phase 5: Every player receives the set of all addresses from root.

    if rank != 0:
        while not timer.expired:
            try:
                raw_message = players[0].next_message(timeout=0.1)
                if raw_message is not None:
                    addresses = pickle.loads(raw_message)
                    log.debug(f"received addresses from player 0.")
                    break
            except ssl.SSLError as e:
                # There was a problem setting up an encrypted connection with
                # the other player, so there's no point in continuing.
                raise EncryptionFailed(message(name, rank, f"failed getting addresses from player 0: {e}"))
            except Exception as e: # pragma: no cover
                log.warning(f"exception getting addresses from player 0: {e}")
                time.sleep(0.1)
        else: # pragma: no cover
            raise Timeout(message(name, rank, "timeout waiting for addresses from player 0."))

    addresses = {rank: (urllib.parse.urlparse(address), token) for rank, (address, token) in addresses.items()}

    ###########################################################################
    # Phase 6: Every player verifies that all tokens match.

    for other_rank, (other_address, other_token) in addresses.items():
        if other_token != token:
            raise TokenMismatch(message(name, rank, f"expected token {token!r}, received {other_token!r} from player {other_rank}."))

    ###########################################################################
    # Phase 7: Players setup connections with one another.

    for listener in range(1, world_size-1):
        if rank == listener:
            # Accept connections from higher-rank players.
            other_players = []
            while not timer.expired:
                if len(other_players) == world_size - rank - 1: # We don't connect to ourself.
                    break

                try:
                    ready = select.select([listen_socket], [], [], 0.1)
                    if ready:
                        sock, _ = listen_socket.accept()
                        if tls is not None:
                            sock = tls[0].wrap_socket(sock, server_side=True)
                            log.info(f"{geturl(sock)} accepted connection from {getpeerurl(sock)}:\n{certformat(sock.getpeercert(), indent='    ')}")
                        else:
                            log.debug(f"{geturl(sock)} accepted connection from {getpeerurl(sock)}.")
                        sock = NetstringSocket(sock)
                        other_players.append(sock)
                except ssl.SSLError as e:
                    # There was a problem setting up an encrypted connection with
                    # the other player, so there's no point in continuing.
                    raise EncryptionFailed(message(name, rank, f"remote player failed to connect: {e}"))
                except Exception as e: # pragma: no cover
                    log.warning(f"exception listening for other players: {type(e)} {e}")
                    time.sleep(0.1)

            # Collect ranks from the other players.
            for other_player in other_players:
                while not timer.expired:
                    try:
                        raw_message = other_player.next_message(timeout=0.1)
                        if raw_message is not None:
                            other_rank = pickle.loads(raw_message)
                            players[other_rank] = other_player
                            log.debug(f"received rank from player {other_rank}.")
                            break
                    except Exception as e: # pragma: no cover
                        log.warning(f"exception receiving player rank: {e}")
                        time.sleep(0.1)
                else: # pragma: no cover
                    raise Timeout(message(name, rank, "timeout waiting for player rank."))

        elif rank > listener:
            # Make a connection to the listener.
            while not timer.expired:
                try:
                    players[listener] = connect(address=addresses[listener][0], rank=rank, other_rank=listener, name=name, tls=tls)
                    break
                except ssl.SSLCertVerificationError as e:
                    # If this happens it means we couldn't verify the other
                    # player's certificate, so there's no point in continuing.
                    raise EncryptionFailed(message(name, rank, f"received invalid certificate from player {listener}."))
                except Exception as e: # pragma: no cover
                    log.warning(f"exception connecting to player {listener}: {e}")
                    time.sleep(0.5)
            else: # pragma: no cover
                raise Timeout(message(name, rank, f"timeout connecting to player {listener}."))

            # Send our rank to the listener.
            while not timer.expired:
                try:
                    players[listener].send(pickle.dumps(rank))
                    break
                except Exception as e: # pragma: no cover
                    log.warning(f"exception sending rank to player {listener}: {e}")
                    time.sleep(0.5)
            else: # pragma: no cover
                raise Timeout(message(name, rank, f"timeout sending rank to player {listener}."))

    return players



# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for creating, manipulating, and revealing additive-shared secrets."""

import numpy

from cicada.additive import AdditiveArrayShare, AdditiveProtocolSuite
from cicada.arithmetic import Field
from cicada.communicator.interface import Communicator
from cicada.encoding import FixedPoint, Identity
from cicada.shamir import ShamirArrayShare, ShamirProtocolSuite


class ActiveArrayShare(object):
    """Stores the local share of a secret shared array for :class:`ActiveProtocolSuite`.

    Instances of :class:`ActiveArrayShare` should only be created
    using :class:`ActiveProtocolSuite`.
    """
    def __init__(self, storage):
        self.storage = storage


    def __repr__(self):
        return f"cicada.active.ActiveArrayShare(storage={self._storage})" # pragma: no cover


    def __getitem__(self, index):
        return ActiveArrayShare((
            AdditiveArrayShare(self.additive.storage[index]),
            ShamirArrayShare(self.shamir.storage[index]))) # pragma: no cover


    @property
    def storage(self):
        """Private storage for the local share of a secret shared array.
        Access is provided only for serialization and communication -
        callers must use :class:`ActiveProtocolSuite` to manipulate secret
        shares.
        """
        return self._storage


    @property
    def additive(self):
        """Additive secret-shared portion of :attr:`storage`.
        """
        return self._storage[0]


    @property
    def shamir(self):
        """Shamir secret-shared portion of :attr:`storage`.
        """
        return self._storage[1]


    @storage.setter
    def storage(self, storage):
        if not len(storage) == 2:
            raise ValueError(f"Expected instances of AdditiveArrayShare and ShamirArrayShare, got {storage} instead.") # pragma: no cover
        if not isinstance(storage[0], AdditiveArrayShare):
            raise ValueError(f"Expected instances of AdditiveArrayShare and ShamirArrayShare, got {storage} instead.") # pragma: no cover
        if not isinstance(storage[1], ShamirArrayShare):
            raise ValueError(f"Expected instances of AdditiveArrayShare and ShamirArrayShare, got {storage} instead.") # pragma: no cover
        self._storage = (storage[0], storage[1])


class ConsistencyError(Exception):
    """Exception raised when the consistency of secret shares cannot be confirmed."""
    pass


class ActiveProtocolSuite(object):
    """Protocol suite implementing shared secret computation that is secure against an active adversary.

    Implements "Combining Shamir & additive secret sharing to improve
    efficiency of SMC primitives against malicious adversaries" by Goss, which
    provides honest majority security with abort.

    Both :class:`~cicada.additive.AdditiveProtocolSuite` and
    :class:`~cicada.shamir.ShamirProtocolSuite` are used in the implementation.

    Note
    ----
    Creating the protocol is a collective operation that *must*
    be called by all players that are members of `communicator`.

    Parameters
    ----------
    communicator: :class:`cicada.communicator.interface.Communicator`, required
        The communicator that this protocol will use for communication.
    threshold: :class:`int`, required
        The minimum number of shares :math:`k` needed to recover a secret,
        where :math:`k <= \\frac{n+1}{2}` and :math:`n` is the number of players.
    seed: :class:`int`, optional
        Seed used to initialize random number generators.  For privacy, this
        value should be different for each player.  By default, the seed will
        be chosen at random, and is guaranteed to be different even on forked
        processes.  If you specify `seed` yourself, the actual seed used will
        be the sum of this value and the value of `seed_offset`.
    seed_offset: :class:`int`, optional
        Value added to the value of `seed`.  This value defaults to the player's
        rank.
    order: :class:`int`, optional
        Field size for storing encoded values.  Defaults to the largest prime
        less than :math:`2^{64}`.
    encoding: :class:`object`, optional
        Default encoding to use for operations that require encoding/decoding.
        Uses an instance of :any:`FixedPoint` with 16 bits of
        floating-point precision if :any:`None`.
    """
    def __init__(self, communicator, *, threshold, seed=None, seed_offset=None, order=None, encoding=None):
        if not isinstance(communicator, Communicator):
            raise ValueError("A Cicada communicator is required.") # pragma: no cover

        if encoding is None:
            encoding = FixedPoint()

        max_threshold = (communicator.world_size+1) // 2
        if threshold > max_threshold:
            min_world_size = (2 * threshold) - 1
            raise ValueError(f"threshold must be <= {max_threshold}, or world_size must be >= {min_world_size}")

        self._communicator = communicator
        self._field = Field(order=order)
        self.aprotocol = AdditiveProtocolSuite(communicator=communicator, seed=seed, seed_offset=seed_offset, order=order, encoding=encoding)
        self.sprotocol = ShamirProtocolSuite(communicator=communicator, threshold=threshold, seed=seed, seed_offset=seed_offset, order=order, encoding=encoding)


    def _assert_binary_compatible(self, lhs, rhs, lhslabel, rhslabel):
        self._assert_unary_compatible(lhs, lhslabel)
        self._assert_unary_compatible(rhs, rhslabel)
        if lhs.additive.storage.shape != rhs.additive.storage.shape :
            raise ValueError(f"{lhslabel} and {rhslabel} additive shares, the additive subshares must be the same shape, got {lhs.additive.storage.shape} and {rhs.additive.storage.shape} instead.") # pragma: no cover
        if lhs.shamir.storage.shape != rhs.shamir.storage.shape:
            raise ValueError(f"{lhslabel} and {rhslabel} shamir shares, the Shamir subshares must be the same shape, got {lhs.shamir.storage.shape} and {rhs.shamir.storage.shape} instead.") # pragma: no cover


    def _assert_unary_compatible(self, share, label):
        if not isinstance(share, ActiveArrayShare):
            raise ValueError(f"{label} must be an instance of ActiveArrayShare, got {type(share)} instead.") # pragma: no cover


    def _require_encoding(self, encoding):
        if encoding is None:
            encoding = self.aprotocol.encoding
        return encoding


    def absolute(self, operand):
        """Elementwise absolute value of a secret shared array.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared value to which the absolute value function should be applied.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared elementwise absolute value of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        return ActiveArrayShare((
            self.aprotocol.absolute(operand.additive),
            self.sprotocol.absolute(operand.shamir)))


    def add(self, lhs, rhs, *, encoding=None):
        """Privacy-preserving elementwise sum of arrays.

        This method can be used to perform private-private, public-private,
        and private-public addition.  The result is the secret shared
        elementwise sum of the operands.  Note that public-public addition
        isn't allowed, as it isn't privacy-preserving!

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public values to be added.
        rhs: :class:`ActiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public values to be added.
        encoding: :class:`object`, optional
            Encodes public operands.  The protocol's :attr:`encoding`
            is used by default if :any:`None`.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared sum of `lhs` and `rhs`.
        """
        encoding = self._require_encoding(encoding)

        # Private-private addition.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, ActiveArrayShare):
            return ActiveArrayShare((
                self.aprotocol.add(lhs.additive, rhs.additive),
                self.sprotocol.add(lhs.shamir, rhs.shamir)))

        # Private-public addition.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, numpy.ndarray):
            return ActiveArrayShare((
                self.aprotocol.add(lhs.additive, rhs),
                self.sprotocol.add(lhs.shamir, rhs)))

        # Public-private addition.
        if isinstance(lhs, numpy.ndarray) and isinstance(rhs, ActiveArrayShare):
            return ActiveArrayShare((
                self.aprotocol.add(lhs, rhs.additive),
                self.sprotocol.add(lhs, rhs.shamir)))

        raise NotImplementedError(f"Privacy-preserving addition not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def bit_compose(self, operand):
        """Compose an array of secret-shared bits into an array of corresponding integer field values.

        The result array will will have one fewer dimensions than the operand,
        which must contain bits in big-endian order in its last dimension.
        Note that the input must contain the field values :math:`0` and :math:`1`.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared array containing bits to be composed.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Share of the resulting field values.
        """
        self._assert_unary_compatible(operand, "operand")
        return ActiveArrayShare((
            self.aprotocol.bit_compose(operand.additive),
            self.sprotocol.bit_compose(operand.shamir)))


    def bit_decompose(self, operand, *, bits=None):
        """Decompose operand into shares of its bitwise representation.

        The result array will have one more dimension than the operand,
        containing the returned bits in big-endian order.  Note that the
        results will contain the field values :math:`0` and :math:`1`.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared array containing values to be decomposed.
        bits: :class:`int`, optional
            The number of rightmost bits in each value to extract.  Defaults to
            all bits (i.e. the number of bits used for storage by the protocol's
            :attr:`field`.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Share of the bit decomposed secret.
        """
        self._assert_unary_compatible(operand, "operand")
        return ActiveArrayShare((
            self.aprotocol.bit_decompose(operand.additive, bits=bits),
            self.sprotocol.bit_decompose(operand.shamir, bits=bits)))


    @property
    def communicator(self):
        """The :class:`~cicada.communicator.interface.Communicator` used by this protocol."""
        return self._communicator


    def divide(self, lhs, rhs, *, encoding=None):
        """Privacy-preserving elementwise division of arrays.

        This method can be used to perform private-private and
        private-public division.  The result is the secret shared
        elementwise quotient of the operands.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared value to be divided.
        rhs: :class:`ActiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be divided.
        encoding: :class:`object`, optional
            Encodes public operands and determines the number of bits to
            shift right from intermediate results.  The protocol's
            :attr:`encoding` is used by default if :any:`None`.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared quotient of `lhs` and `rhs`.
        """
        encoding = self._require_encoding(encoding)

        # Private-private division.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, ActiveArrayShare):
            truncbits = encoding.precision
            tbm, tshare = self.random_bitwise_secret(bits=truncbits, shape=rhs.additive.storage.shape)

            rhsmasked = self.field_multiply(tshare, rhs)
            rhsmasked = self.right_shift(rhsmasked, bits=encoding.precision)
            revealrhsmasked = self.reveal(rhsmasked, encoding=encoding)
            almost_there = self.right_shift(self.field_multiply(lhs, tshare), bits=encoding.precision)
            divisor = encoding.encode(numpy.array(1/revealrhsmasked), self.field)
            quotient = self.field_multiply(almost_there, divisor)
            return self.right_shift(quotient, bits=encoding.precision)

        # Private-public division.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, numpy.ndarray):
            return ActiveArrayShare((
                self.aprotocol.divide(lhs.additive, rhs, encoding=encoding),
                self.sprotocol.divide(lhs.shamir, rhs, encoding=encoding)))

        # Public-private division.
        if isinstance(lhs, numpy.ndarray) and isinstance(rhs, ActiveArrayShare):
            pass

        raise NotImplementedError(f"Privacy-preserving division not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def dot(self, lhs, rhs, *, encoding=None):
        """Privacy-preserving dot product of two secret shared vectors.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared vector.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared vector.
        encoding: :class:`object`, optional
            Determines the number of bits to truncate from intermediate
            results.  The protocol's :attr:`encoding` is used by default if
            :any:`None`.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared dot product of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((
            self.aprotocol.dot(lhs.additive, rhs.additive, encoding=encoding),
            self.sprotocol.dot(lhs.shamir, rhs.shamir, encoding=encoding)))


    def equal(self, lhs, rhs):
        """Elementwise probabilistic equality comparison between secret shared arrays.

        The result is the secret shared elementwise comparison `lhs` == `rhs`.  Note
        that the results will contain the field values :math:`0` and :math:`1`.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared array to be compared.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared array to be compared.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared result from comparing `lhs` == `rhs` elementwise.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((
            self.aprotocol.equal(lhs.additive, rhs.additive),
            self.sprotocol.equal(lhs.shamir, rhs.shamir)))


    @property
    def field(self):
        """Integer :any:`Field` used for arithmetic on and storage of secret shared values."""
        return self._field


    def field_add(self, lhs, rhs):
        """Privacy-preserving elementwise sum of arrays.

        This method can be used to perform private-private, public-private,
        and private-public addition.  The result is the secret shared
        elementwise sum of the operands.  Note that public-public addition
        isn't allowed, as it isn't privacy-preserving!

        Unlike :meth:`add`, :meth:`field_add` only operates on field
        values, no encoding is performed on its inputs.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be added.
        rhs: :class:`ActiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be added.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared sum of `lhs` and `rhs`.
        """
        # Private-private addition.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, ActiveArrayShare):
            return ActiveArrayShare((
                self.aprotocol.field_add(lhs.additive, rhs.additive),
                self.sprotocol.field_add(lhs.shamir, rhs.shamir)))

        # Private-public addition.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, numpy.ndarray):
            return ActiveArrayShare((
                self.aprotocol.field_add(lhs.additive, rhs),
                self.sprotocol.field_add(lhs.shamir, rhs)))

        # Public-private addition.
        if isinstance(lhs, numpy.ndarray) and isinstance(rhs, ActiveArrayShare):
            return ActiveArrayShare((
                self.aprotocol.field_add(lhs, rhs.additive),
                self.sprotocol.field_add(lhs, rhs.shamir)))

        raise NotImplementedError(f"Privacy-preserving addition not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def field_dot(self, lhs, rhs):
        """Privacy-preserving dot product of two secret shared vectors.

        Unlike :meth:`dot`, :meth:`field_dot` only operates on field values,
        no right shift is performed on the results.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared vector.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared vector.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared dot product of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((
            self.aprotocol.field_dot(lhs.additive, rhs.additive),
            self.sprotocol.field_dot(lhs.shamir, rhs.shamir)))


    def field_multiply(self, lhs, rhs):
        """Privacy-preserving elementwise multiplication of arrays.

        This method can be used to perform private-private, public-private, and
        private-public multiplication.  The result is the secret shared
        elementwise sum of the operands.  Note that public-public
        multiplication isn't allowed, as it isn't privacy-preserving!

        Unlike :meth:`multiply`, :meth:`field_multiply` only operates on field
        values, no encoding is performed on its inputs, and no right shift
        is performed on the results.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be multiplied.
        rhs: :class:`ActiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be multiplied.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared product of `lhs` and `rhs`.
        """
        # Private-private multiplication.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, ActiveArrayShare):
            return ActiveArrayShare((
                self.aprotocol.field_multiply(lhs.additive, rhs.additive),
                self.sprotocol.field_multiply(lhs.shamir, rhs.shamir)))

        # Public-private multiplication.
        if isinstance(lhs, numpy.ndarray) and isinstance(rhs, ActiveArrayShare):
            return ActiveArrayShare((
                self.aprotocol.field_multiply(lhs, rhs.additive),
                self.sprotocol.field_multiply(lhs, rhs.shamir)))

        # Private-public multiplication.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, numpy.ndarray):
            return ActiveArrayShare((
                self.aprotocol.field_multiply(lhs.additive, rhs),
                self.sprotocol.field_multiply(lhs.shamir, rhs)))

        raise NotImplementedError(f"Privacy-preserving multiplication not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def field_power(self, lhs, rhs):
        """Privacy-preserving elementwise exponentiation.

        Raises secret shared array values to public integer values.  Unlike :meth:`power`,
        :meth:`field_power` only operates on field values, no right shift is performed on the
        results.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared values which iwll be raised to a power.
        rhs: :class:`int` or integer :class:`numpy.ndarray`, required
            Public integer power(s) to which each element in `lhs` will be raised.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared result of raising `lhs` to the power(s) in `rhs`.
        """
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, (numpy.ndarray, int)):
            if isinstance(rhs, int):
                rhs = self.field.full_like(lhs.additive.storage, rhs)
            return ActiveArrayShare((
                self.aprotocol.field_power(lhs.additive, rhs),
                self.sprotocol.field_power(lhs.shamir, rhs)))

        raise NotImplementedError(f"Privacy-preserving exponentiation not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def field_subtract(self, lhs, rhs):
        """Privacy-preserving elementwise difference of arrays.

        This method can be used to perform private-private, public-private, and
        private-public subtraction.  The result is the secret shared
        elementwise difference of the operands.  Note that public-public
        subtraction isn't allowed, as it isn't privacy-preserving!

        Unlike :meth:`subtract`, :meth:`field_subtract` only operates on field
        values, no encoding is performed on its inputs.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be subtracted.
        rhs: :class:`ActiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be subtracted.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared difference of `lhs` and `rhs`.
        """
        # Private-private subtraction.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, ActiveArrayShare):
            return ActiveArrayShare((
                self.aprotocol.field_subtract(lhs.additive, rhs.additive),
                self.sprotocol.field_subtract(lhs.shamir, rhs.shamir)))

        # Private-public subtraction.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, numpy.ndarray):
            return ActiveArrayShare((
                self.aprotocol.field_subtract(lhs.additive, rhs),
                self.sprotocol.field_subtract(lhs.shamir, rhs)))

        # Public-private subtraction.
        if isinstance(lhs, numpy.ndarray) and isinstance(rhs, ActiveArrayShare):
            return ActiveArrayShare((
                self.aprotocol.field_subtract(lhs, rhs.additive),
                self.sprotocol.field_subtract(lhs, rhs.shamir)))

        raise NotImplementedError(f"Privacy-preserving subtraction not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def field_uniform(self, *, shape=None, generator=None):
        """Generate private random field elements.

        This method can be used to generate a secret shared array containing
        random field elements of any shape.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        shape: :class:`tuple`, optional
            Shape of the array to populate. By default,
            a shapeless array of one random element will be generated.
        src: sequence of :class:`int`, optional
            Players that will contribute to random array generation.  By default,
            all players contribute.
        generator: :class:`numpy.random.Generator`, optional
            A psuedorandom number generator for sampling.  By default,
            `numpy.random.default_rng()` will be used.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared array of random field elements.
        """
        uniadd = self.aprotocol.field_uniform(shape=shape, generator=generator)
        shamadd = []
        for i in self.communicator.ranks:
            shamadd.append(self.sprotocol.share(src=i, secret=uniadd.storage, shape=uniadd.storage.shape, encoding=Identity()))
        unisham = ShamirArrayShare(numpy.array(sum([x.storage for x in shamadd]), dtype=self.field.dtype))
        return ActiveArrayShare((uniadd, unisham))


    def floor(self, operand, *, encoding=None):
        """Privacy-preserving elementwise floor of encoded, secret-shared arrays.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared values to which floor should be applied.
        encoding: :class:`object`, optional
            Determines the number of fractional bits used for encoded values.
            The protocol's :attr:`encoding` is used by default if :any:`None`.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared floor of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        return ActiveArrayShare((
            self.aprotocol.floor(operand.additive, encoding=encoding),
            self.sprotocol.floor(operand.shamir, encoding=encoding)))


    def less(self, lhs, rhs):
        r"""Privacy-preserving elementwise less-than comparison.

        The result is the secret shared elementwise comparison :math:`lhs \lt rhs`.
        Note that the results will contain the field values :math:`0` or
        :math:`1`, which do not require decoding if revealed.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared values to be compared.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared values to be compared.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared comparison :math:`lhs \lt rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((
            self.aprotocol.less(lhs.additive, rhs.additive),
            self.sprotocol.less(lhs.shamir, rhs.shamir)))


    def less_zero(self, operand):
        r"""Privacy-preserving elementwise less-than-zero comparison.

        The result is the secret shared elementwise comparison :math:`operand \lt 0`.
        Note that the results will contain the field values :math:`0` or
        :math:`1`, which do not require decoding if revealed.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared values to be compared.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared values to be compared.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared comparison :math:`operand \lt 0`.
        """
        self._assert_unary_compatible(operand, "operand")
        return ActiveArrayShare((
            self.aprotocol.less_zero(operand.additive),
            self.sprotocol.less_zero(operand.shamir)))


    def logical_and(self, lhs, rhs):
        """Privacy-preserving elementwise logical AND of secret shared arrays.

        The operands *must* contain the field values :math:`0` or :math:`1`.
        The result will be the secret shared elementwise logical AND of `lhs`
        and `rhs`, and will also contain the field values :math:`0` or
        :math:`1`, which do not require decoding if revealed.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared array for logical AND.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared array for logical AND.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared elementwise logical AND of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((
            self.aprotocol.logical_and(lhs.additive, rhs.additive),
            self.sprotocol.logical_and(lhs.shamir, rhs.shamir)))


    def logical_not(self, operand):
        """Privacy-preserving elementwise logical NOT.

        The operand *must* contain the field values :math:`0` or :math:`1`.
        The result will be the secret shared elementwise logical NOT of
        `operand`, and will also contain the field values :math:`0` or
        :math:`1`, which do not require decoding if revealed.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared array for logical NOT.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared elementwise logical NOT of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        return ActiveArrayShare((
            self.aprotocol.logical_not(operand.additive),
            self.sprotocol.logical_not(operand.shamir)))


    def logical_or(self, lhs, rhs):
        """Privacy-preserving elementwise logical OR of secret shared arrays.

        The operands *must* contain the field values :math:`0` or :math:`1`.
        The result will be the secret shared elementwise logical OR of `lhs`
        and `rhs`, and will also contain the field values :math:`0` or
        :math:`1`, which do not require decoding if revealed.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared array for logical OR.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared array for logical OR.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared elementwise logical OR of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((
            self.aprotocol.logical_or(lhs.additive, rhs.additive),
            self.sprotocol.logical_or(lhs.shamir, rhs.shamir)))


    def logical_xor(self, lhs, rhs):
        """Privacy-preserving elementwise logical XOR of secret shared arrays.

        The operands *must* contain the field values :math:`0` or :math:`1`.
        The result will be the secret shared elementwise logical XOR of `lhs`
        and `rhs`, and will also contain the field values :math:`0` or
        :math:`1`, which do not require decoding if revealed.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared array for logical XOR.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared array for logical XOR.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared elementwise logical XOR of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((
            self.aprotocol.logical_xor(lhs.additive, rhs.additive),
            self.sprotocol.logical_xor(lhs.shamir, rhs.shamir)))


    def maximum(self, lhs, rhs):
        """Privacy-preserving elementwise maximum of secret shared arrays.

        The result is the secret shared elementwise maximum of the operands.
        Note: the magnitude of the field elements should be less than one
        quarter of the field order for this method to be accurate in general.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared operand.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared operand.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared elementwise maximum of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((
            self.aprotocol.maximum(lhs.additive, rhs.additive),
            self.sprotocol.maximum(lhs.shamir, rhs.shamir)))


    def minimum(self, lhs, rhs):
        """Privacy-preserving elementwise minimum of secret shared arrays.

        The result is the secret shared elementwise minimum of the operands.
        Note: the magnitude of the field elements should be less than one
        quarter of the field order for this method to be accurate in general.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared operand.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared operand.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared elementwise minimum of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((
            self.aprotocol.minimum(lhs.additive, rhs.additive),
            self.sprotocol.minimum(lhs.shamir, rhs.shamir)))


    def multiplicative_inverse(self, operand):
        """Privacy-preserving elementwise multiplicative inverse of a secret shared array.

        Returns the multiplicative inverse of a secret shared array
        in the context of the underlying finite field. Explicitly, this
        function returns a same shape array which, when multiplied
        elementwise with `operand`, will return a same shape array comprised
        entirely of ones, assuming `operand` is entirely non-trivial elements.

        This function does not take into account any field-external symantics.
        There is a potential for information leak here if `operand` contains any
        zero elements, that will be revealed. There is a small probability,
        2^-operand.storage.size, for this approach to fail by zero being randomly
        generated by the parties as the mask.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared operand to be multiplicatively inverted.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared elementwise multiplicative inverse of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        return ActiveArrayShare((
            self.aprotocol.multiplicative_inverse(operand.additive),
            self.sprotocol.multiplicative_inverse(operand.shamir)))


    def multiply(self, lhs, rhs, *, encoding=None):
        """Privacy-preserving elementwise multiplication of arrays.

        This method can be used to perform private-private, public-private, and
        private-public multiplication.  The result is the secret shared
        elementwise sum of the operands.  Note that public-public
        multiplication isn't allowed, as it isn't privacy-preserving!

        Unlike :meth:`field_multiply`, :meth:`multiply` is encoding aware:
        encoding is performed on its public inputs, and the results are
        shifted right to produce correct results when decoded.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be multiplied.
        rhs: :class:`ActiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be multiplied.
        encoding: :class:`object`, optional
            Encodes public operands and determines the number of bits to shift
            right the results.  The protocol's :attr:`encoding` is used by
            default if :any:`None`.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared product of `lhs` and `rhs`.
        """
        # Private-private multiplication.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, ActiveArrayShare):
            return ActiveArrayShare((
                self.aprotocol.multiply(lhs.additive, rhs.additive, encoding=encoding),
                self.sprotocol.multiply(lhs.shamir, rhs.shamir, encoding=encoding)))

        # Private-public multiplication.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, numpy.ndarray):
            return ActiveArrayShare((
                self.aprotocol.multiply(lhs.additive, rhs, encoding=encoding),
                self.sprotocol.multiply(lhs.shamir, rhs, encoding=encoding)))

        # Public-private multiplication.
        if isinstance(lhs, numpy.ndarray) and isinstance(rhs, ActiveArrayShare):
            return ActiveArrayShare((
                self.aprotocol.multiply(lhs, rhs.additive, encoding=encoding),
                self.sprotocol.multiply(lhs, rhs.shamir, encoding=encoding)))

        raise NotImplementedError(f"Privacy-preserving multiplication not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def negative(self, operand):
        """Privacy-preserving elementwise additive inverse of a secret shared array.

        Returns the additive inverse of a secret shared array
        in the context of the underlying finite field. Explicitly, this
        function returns a same shape array which, when added
        elementwise with `operand`, will return a same shape array comprised
        entirely of zeros.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared operand to be additively inverted.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared elementwise additive inverse of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        return ActiveArrayShare((
            self.aprotocol.negative(operand.additive),
            self.sprotocol.negative(operand.shamir)))


#    def _pade_approx(self, func, operand,*, encoding=None, center=0, degree=12, scale=3):
#        """Return the pade approximation of `func` sampled with `operand`.
#
#        Note
#        ----
#        This is a collective operation that *must* be called
#        by all players that are members of :attr:`communicator`.
#
#        Parameters
#        ----------
#        func: callable object, required
#            The function to be approximated via the pade method.
#        operand: :class:`ActiveArrayShare`, required
#            Secret-shared values where `func` should be evaluated.
#        center: :class:`float`, optional
#            The value at which the approximation should be centered. Sample
#            errors will be larger the further they are from this point.
#
#        Returns
#        -------
#        result: :class:`ActiveArrayShare`
#            Secret shared result of evaluating the pade approximant of func(operand) with the given parameters.
#        """
#        from scipy.interpolate import approximate_taylor_polynomial, pade
#        num_deg = degree%2+degree//2
#        den_deg = degree//2
#
#        self._assert_unary_compatible(operand, "operand")
#        encoding = self._require_encoding(encoding)
#
#        func_taylor = approximate_taylor_polynomial(func, center, degree, scale)
#        func_pade_num, func_pade_den = pade([x for x in func_taylor][::-1], den_deg, n=num_deg)
#        enc_func_pade_num = encoding.encode(numpy.array([x for x in func_pade_num]), self.field)
#        enc_func_pade_den = encoding.encode(numpy.array([x for x in func_pade_den]), self.field)
#        op_pows_num_list = [self.share(src=1, secret=numpy.array(1), shape=())]
#        for i in range(num_deg):
#            op_pows_num_list.append(self.multiply(operand, op_pows_num_list[-1]))
#        if degree%2:
#            op_pows_den_list=[thing for thing in op_pows_num_list[:-1]]
#        else:
#            op_pows_den_list=[thing for thing in op_pows_num_list]
#
#        num_add_shares = AdditiveArrayShare(numpy.array([x.additive.storage for x in op_pows_num_list]))
#        num_sham_shares = ShamirArrayShare(numpy.array([x.shamir.storage for x in op_pows_num_list]))
#        den_add_shares = AdditiveArrayShare(numpy.array([x.additive.storage for x in op_pows_den_list]))
#        den_sham_shares = ShamirArrayShare(numpy.array([x.shamir.storage for x in op_pows_den_list]))
#        op_pows_num = ActiveArrayShare((num_add_shares, num_sham_shares))
#        op_pows_den = ActiveArrayShare((den_add_shares, den_sham_shares ))
#
#        result_num_prod = self.field_multiply(op_pows_num, enc_func_pade_num)
#        result_num = self.right_shift(self.sum(result_num_prod), bits=encoding.precision)
#
#        result_den_prod = self.field_multiply(op_pows_den, enc_func_pade_den)
#        result_den = self.right_shift(self.sum(result_den_prod), bits=encoding.precision)
#        result = self.divide(result_num, result_den)
#        return result



    def power(self, lhs, rhs, *, encoding=None):
        """Privacy-preserving elementwise exponentiation.

        Raises secret shared array values to public integer values.  Unlike
        :meth:`field_power`, :meth:`power` operates on encoded values, shifting
        the results right to ensure correct decoded results.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared values which iwll be raised to a power.
        rhs: :class:`int` or integer :class:`numpy.ndarray`, required
            Public integer power(s) to which each element in `lhs` will be raised.
        encoding: :class:`object`, optional
            Determines the number of bits to shift right the results.  The
            protocol's :attr:`encoding` is used by default if :any:`None`.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared result of raising `lhs` to the power(s) in `rhs`.
        """
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, (numpy.ndarray, int)):
            if isinstance(rhs, int):
                rhs = self.field.full_like(lhs.storage, rhs)
            return ActiveArrayShare((
                self.aprotocol.power(lhs.additive, rhs),
                self.sprotocol.power(lhs.shamir, rhs)))

        raise NotImplementedError(f"Privacy-preserving exponentiation not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def random_bitwise_secret(self, *, bits, src=None, generator=None, shape=None):
        """Return secret values created by combining randomly generated bits.

        This method returns two outputs - a secret shared array of randomly
        generated bits, and a secret shared array of values created by
        combining the bits in big-endian order.  It is secure against
        non-colluding semi-honest adversaries.  A subset of players (by
        default: all) generate and secret share vectors of pseudo-random bits
        which are then XOR-ed together elementwise.  Communication and
        computation costs increase with the number of bits and the number of
        players, while security increases with the number of players.

        The bit array will have one more dimension than the secret array,
        containing the bits in big-endian order.

        .. warning::

             If you supply your own generators, be careful to ensure that each
             has a unique seed value to preserve privacy (for example: a
             constant plus the player's rank).  If players receive generators
             with identical seed values, even numbers of players will produce
             all zero bits.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`, even
        if they aren't participating in the random bit generation.

        Parameters
        ----------
        bits: :class:`int`, required
            Number of bits to generate.
        shape: sequence of :class:`int`, optional
            Shape of the output `secrets` array.  The output `bits` array will have this
            shape, plus one extra dimension for the bits.
        src: sequence of :class:`int`, optional
            Players that will contribute to random bit generation.  By default,
            all players contribute.
        generator: :class:`numpy.random.Generator`, optional
            A psuedorandom number generator for sampling.  By default,
            `numpy.random.default_rng()` will be used.

        Returns
        -------
        bits: :class:`ActiveArrayShare`
            Secret shared array of randomly-generated bits, with shape
            :math:`shape \\times bits`.
        secrets: :class:`ActiveArrayShare`
            Secret shared array of values created by combining the generated
            bits in big-endian order, with shape `shape`.
        """
        bs_add, ss_add = self.aprotocol.random_bitwise_secret(bits=bits, src=src, generator=generator, shape=shape)
        sham_bs_add = []
        sham_assembled_add = []
        for i in self.communicator.ranks:
            sham_bs_add.append(self.sprotocol.share(src=i, secret=bs_add.storage, shape=bs_add.storage.shape, encoding=Identity()))
            sham_assembled_add.append(self.sprotocol.share(src=i, secret=ss_add.storage, shape=ss_add.storage.shape, encoding=Identity()))
        bs_sham = ShamirArrayShare(numpy.array(sum([x.storage for x in sham_bs_add])%self._field.order, dtype=self.field.dtype))
        ss_sham = ShamirArrayShare(numpy.array(sum([x.storage for x in sham_assembled_add])%self._field.order, dtype=self.field.dtype))
        bs_active = ActiveArrayShare((bs_add, bs_sham))
        assembled_active = ActiveArrayShare((ss_add, ss_sham))
        return (bs_active, assembled_active)


    def relu(self, operand):
        """Privacy-preserving elementwise ReLU of a secret shared array.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared operand to which the ReLU function will be applied.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared elementwise ReLU of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        return ActiveArrayShare((
            self.aprotocol.relu(operand.additive),
            self.sprotocol.relu(operand.shamir)))


    def reshare(self, operand):
        """Privacy-preserving re-randomization of a secret shared array.

        This method returns a new set of secret shares that contain different
        random values than the input but represent the same secret value.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared operand which should be re-randomized.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared, re-randomized version of `operand`.

        Raises
        ------
        :class:`ConsistencyError`
            If the consistency of the returned secret shares cannot be confirmed.
        """
        self._assert_unary_compatible(operand, "operand")

        reshared = ActiveArrayShare((
            self.aprotocol.reshare(operand.additive),
            self.sprotocol.reshare(operand.shamir)))

        if not self.verify(reshared):
            raise ConsistencyError("Secret Shares being reshared are inconsistent") # pragma: no cover

        return reshared


    def reveal(self, share, *, dst=None, encoding=None):
        """Reveals a secret shared value to a subset of players.

        Note
        ----
        This is a collective operation that *must* be called by all players
        that are members of :attr:`communicator`, whether they are receiving
        the revealed secret or not.

        Parameters
        ----------
        share: :class:`ActiveArrayShare`, required
            The local share of the secret to be revealed.
        dst: sequence of :class:`int`, optional
            List of players who will receive the revealed secret.  If :any:`None`
            (the default), the secret will be revealed to all players.
        encoding: :class:`object`, optional
            Encoding used to extract the revealed secret from field values. The
            protocol's default encoding will be used if `None`.

        Returns
        -------
        value: :class:`numpy.ndarray` or :any:`None`
            The revealed secret, if this player is a member of `dst`, or :any:`None`.

        Raises
        ------
        :class:`ConsistencyError`
            If the consistency of the returned secret shares cannot be confirmed.
        """
        if not isinstance(share, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(share)} instead.") # pragma: no cover

        # Identify who will be receiving shares.
        if dst is None:
            dst = self.communicator.ranks

        encoding = self._require_encoding(encoding)

        zshare = ShamirArrayShare(self.sprotocol.field.subtract(share.shamir.storage, numpy.array((pow(self.sprotocol._revealing_coef[self.communicator.rank], self.field.order-2, self.field.order) * share.additive.storage) % self.field.order, dtype=self.field.dtype)))

        a_storage = numpy.array(self.communicator.allgather(share.additive.storage), dtype=self.field.dtype)
        z_storage = numpy.array(self.communicator.allgather(zshare.storage), dtype=self.field.dtype)
        secret = []
        revealing_coef = self.sprotocol._revealing_coef
        for index in numpy.ndindex(z_storage[0].shape):
            secret.append(sum([revealing_coef[i]*z_storage[i][index] for i in range(len(revealing_coef))]))
        rev = numpy.array([x%self.field.order for x in secret], dtype=self.field.dtype).reshape(share.additive.storage.shape)
        if len(rev.shape) == 0 and rev:
            raise ConsistencyError("Secret Shares are inconsistent in the first stage") # pragma: no cover
        if len(rev.shape) > 0 and numpy.any(rev):
            raise ConsistencyError("Secret Shares are inconsistent in the first stage") # pragma: no cover

        secret = []
        for index in numpy.ndindex(a_storage[0].shape):
            secret.append(sum([a_storage[i][index] for i in range(len(revealing_coef))]))
        reva = numpy.array([x%self.field.order for x in secret], dtype=self.field.dtype).reshape(share.additive.storage.shape)
        bs_storage=numpy.zeros(z_storage.shape, dtype=self.field.dtype)
        for i, c in enumerate(revealing_coef):
            bs_storage[i] =  self.sprotocol.field.add(z_storage[i], numpy.array((pow(self.sprotocol._revealing_coef[i], self.field.order-2, self.field.order) * a_storage[i]) % self.field.order, dtype=self.field.dtype))
        bs_storage %= self.field.order
        s1 = numpy.sort(numpy.random.choice(self.sprotocol.indices, self.sprotocol._d+1, replace=False))
        revealing_coef = self.sprotocol._lagrange_coef(s1)
        sub_secret = []
        if len(z_storage[0].shape) > 0:
            for index in numpy.ndindex(z_storage[0].shape):
                loop_acc = 0
                for i,v in enumerate(s1):
                    loop_acc += revealing_coef[i]*bs_storage[v-1][index]
                sub_secret.append(loop_acc % self.field.order)
        else:
            loop_acc = 0
            for i,v in enumerate(s1):
                loop_acc += revealing_coef[i]*bs_storage[v-1]
            sub_secret.append(loop_acc % self.field.order)

        revs = numpy.array(sub_secret, dtype=self.field.dtype).reshape(share.additive.storage.shape)
        while True:
            s2 = numpy.sort(numpy.random.choice(self.sprotocol.indices, self.sprotocol._d+1, replace=False))
            if not numpy.array_equal(s1, s2):
                break
        revealing_coef = self.sprotocol._lagrange_coef(s2)
        sub_secret2 = []
        if len(z_storage[0].shape) > 0:
            for index in numpy.ndindex(z_storage[0].shape):
                loop_acc = 0
                for i,v in enumerate(s2):
                    loop_acc += revealing_coef[i]*bs_storage[v-1][index]
                sub_secret2.append(loop_acc % self.field.order)
        else:
            loop_acc = 0
            for i,v in enumerate(s2):
                loop_acc += revealing_coef[i]*bs_storage[v-1]
            sub_secret2.append(loop_acc % self.field.order)

        revs2 = numpy.array(sub_secret2, dtype=self.field.dtype).reshape(share.additive.storage.shape)
        if len(revs.shape) > 0 or len(revs2.shape) > 0:
            if numpy.any(revs != reva) or numpy.any(revs2 != reva):
                raise ConsistencyError("Secret Shares are inconsistent in the second stage") # pragma: no cover
        else:
            if revs != reva or revs2 != reva:
                raise ConsistencyError("Secret Shares are inconsistent in the second stage") # pragma: no cover

        return encoding.decode(revs, self.field)


    def right_shift(self, operand, *, bits, src=None, generator=None):
        """Privacy-preserving elementwise right-shift of a secret shared array.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Also, this approach may probabilstically introduce a small error in the least significant bits 
        of the result. The is that we generate masks for the two sections of the secret shared value 
        of interest, one to be used in clearing out the low order bits, the other for masking the 
        higher order bits that will remain. There is a chance that adding the masks in will generate 
        a carry into the region that will remain after the shift. Therefore there is a chance that the 
        effects of that carry will remain in the least significant bits of the final result, which, 
        with the default parameters and encoding, means there could be an error in the final result on 
        the order of 2^-16, and with decreasing probability 2^-15, or 2^-14 depending on how far the 
        carry propagates.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret-shared values to be shifted right.
        bits: :class:`int`, optional
            Number of bits to shift.
        src: sequence of :class:`int`, optional
            Players who will participate in generating random bits as part of
            the shift process.  More players increases security but
            decreases performance.  Defaults to all players.
        generator: :class:`numpy.random.Generator`, optional
            A psuedorandom number generator for generating random bits.  By
            default, `numpy.random.default_rng()` will be used.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared result of shifting `operand` to the right by `bits` bits.
        """
        if not isinstance(operand, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover

        tbm, tshare = self.random_bitwise_secret(bits=bits, src=src, generator=generator, shape=operand.additive.storage.shape)
        rbm, rshare = self.random_bitwise_secret(bits=self.field.bits-bits, src=src, generator=generator, shape=operand.additive.storage.shape)
        return ActiveArrayShare((
            self.aprotocol.right_shift(operand.additive, bits=bits, src=src, generator=generator, trunc_mask=tshare.additive, rem_mask=rshare.additive),
            self.sprotocol.right_shift(operand.shamir, bits=bits, src=src, generator=generator, trunc_mask=tshare.shamir, rem_mask=rshare.shamir)))


    def share(self, *, src, secret, shape, encoding=None):
        """Convert an array of scalars to an additive secret share.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        src: :class:`int`, required
            The player providing the private array to be secret shared.
        secret: :class:`numpy.ndarray` or :any:`None`, required
            The secret array to be shared.  This value is ignored for all
            players except `src`.
        shape: :class:`tuple`, required
            The shape of the secret.  Note that all players must specify the
            same shape.
        encoding: :class:`object`, optional
            Encoding used to convert `secret` into field values. The protocol's
            default encoding will be used if `None`.

        Returns
        -------
        share: :class:`ActiveArrayShare`
            The local share of the secret shared array.
        """
        return ActiveArrayShare((
            self.aprotocol.share(src=src, secret=secret, shape=shape, encoding=encoding),
            self.sprotocol.share(src=src, secret=secret, shape=shape, encoding=encoding)))


    def subtract(self, lhs, rhs, *, encoding=None):
        """Privacy-preserving elementwise difference of arrays.

        This method can be used to perform private-private, public-private, and
        private-public addition.  The result is the secret shared elementwise
        difference of the operands.  Note that public-public subtraction isn't
        allowed, as it isn't privacy-preserving!

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public values to be subtracted.
        rhs: :class:`ActiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public values to be subtracted.
        encoding: :class:`object`, optional
            Encodes public operands.  The protocol's :attr:`encoding`
            is used by default if :any:`None`.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared difference of `lhs` and `rhs`.
        """
        encoding = self._require_encoding(encoding)

        # Private-private subtraction.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, ActiveArrayShare):
            return ActiveArrayShare((
                self.aprotocol.subtract(lhs.additive, rhs.additive),
                self.sprotocol.subtract(lhs.shamir, rhs.shamir)))

        # Private-public subtraction.
        if isinstance(lhs, ActiveArrayShare) and isinstance(rhs, numpy.ndarray):
            return ActiveArrayShare((
                self.aprotocol.subtract(lhs.additive, rhs),
                self.sprotocol.subtract(lhs.shamir, rhs)))

        # Public-private subtraction.
        if isinstance(lhs, numpy.ndarray) and isinstance(rhs, ActiveArrayShare):
            return ActiveArrayShare((
                self.aprotocol.subtract(lhs, rhs.additive),
                self.sprotocol.subtract(lhs, rhs.shamir)))

        raise NotImplementedError(f"Privacy-preserving subtraction not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def sum(self, operand):
        """Privacy-preserving sum of a secret shared array's elements.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared array containing elements to be summed.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            Secret-shared sum of `operand`'s elements.
        """
        self._assert_unary_compatible(operand, "operand")
        return ActiveArrayShare((
            self.aprotocol.sum(operand.additive),
            self.sprotocol.sum(operand.shamir)))


    def verify(self, operand):
        """Test whether shares are consistent.

        Tests the consistency of secret shares among all players using a zero
        knowledge proof. This method allows the players to prove to one another
        (in zero knowledge) that a consistent set of shares is known by all
        players.  It can provide a "safe" point at which calculations are known
        to be good, so that parties can "rewind" if consistency problems are
        later discovered in the protocol.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared operand which will be verified for consistency.

        Returns
        -------
        consistency: :class:`bool`
            :any:`True` if every player has demonstrated that they have consistent shares, otherwise :any:`False`.
        """
        self._assert_unary_compatible(operand, "operand")

        a_share = operand.additive
        s_share = operand.shamir
        zero = ShamirArrayShare(self.sprotocol.field.subtract(s_share.storage, numpy.array((pow(self.sprotocol._revealing_coef[self.communicator.rank], self.field.order-2, self.field.order) * a_share.storage) % self.field.order, dtype=object)))
        consistency = numpy.all(self.sprotocol.reveal(zero, encoding=Identity()) == numpy.zeros(zero.storage.shape))
        return consistency



#    def _taylor_approx(self, func, operand,*, encoding=None, center=0, degree=7, scale=3):
#        """Return the taylor approximation of `func` sampled with `operand`.
#
#        Note
#        ----
#        This is a collective operation that *must* be called
#        by all players that are members of :attr:`communicator`.
#
#        Parameters
#        ----------
#        func: callable object, required
#            The function to be approximated via the taylor method
#        operand: :class:`ActiveArrayShare`, required
#            Secret-shared values where `func` should be evaluated.
#        center: :class:`float`, optional
#            The value at which the approximation should be centered. Sample
#            errors will be larger the further they are from this point.
#
#        Returns
#        -------
#        result: :class:`ActiveArrayShare`
#            Secret shared result of evaluating the taylor approximant of func(operand) with the given parameters
#        """
#        from scipy.interpolate import approximate_taylor_polynomial
#
#        self._assert_unary_compatible(operand, "operand")
#        encoding = self._require_encoding(encoding)
#
#        taylor_poly = approximate_taylor_polynomial(func, center, degree, scale)
#
#        enc_taylor_coef = encoding.encode(numpy.array([x for x in taylor_poly]), self.field)
#        op_pow_list = [self.share(src=1, secret=numpy.array(1), shape=())]
#        for i in range(degree):
#            op_pow_list.append(self.multiply(operand, op_pow_list[-1]))
#
#        op_pow_add_shares = AdditiveArrayShare(numpy.array([x.additive.storage for x in op_pow_list]))
#        op_pow_sham_shares = ShamirArrayShare(numpy.array([x.shamir.storage for x in op_pow_list]))
#        op_pow_shares = ActiveArrayShare((op_pow_add_shares, op_pow_sham_shares))
#
#        result = self.field_multiply(op_pow_shares, enc_taylor_coef)
#        result = self.sum(result)
#        result = self.right_shift(result, bits=encoding.precision)
#        return result


    def zigmoid(self, operand, *, encoding=None):
        r"""Privacy-preserving elementwise zigmoid of a secret shared array.

        Zigmoid is a piecewise approximation to the popular sigmoid activation
        function that is more efficient to compute in an MPC context:

        .. math::

            zigmoid(x) = \left\{
            \begin{array}\\
                0 & if\ x<-0.5 \\
                x+0.5 & if\ -0.5\leq x \leq 0.5 \\
                1 & if x>0.5
            \end{array}
            \right.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared operand to which the zigmoid function will be applied.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared elementwise zigmoid of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        return ActiveArrayShare((
            self.aprotocol.zigmoid(operand.additive, encoding=encoding),
            self.sprotocol.zigmoid(operand.shamir, encoding=encoding)))




# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for generating transcripts of library activity.

Use-cases include detailed debugging, logging network traffic, function
tracing, and MPC-in-the-head for zero knowledge proofs.  The latter case
is described in

    Ishai, Yuval, et al. "Zero-knowledge from secure multiparty computation." *Proceedings of the thirty-ninth annual ACM symposium on Theory of computing. 2007.*

Examples of this technique appear in
https://csrc.nist.gov/projects/pqc-dig-sig/round-1-additional-signatures as
proposals for NIST PQC standardization.
"""

import copy
import logging
import types

import hunter
import numpy

from .active import ActiveProtocolSuite
from .additive import AdditiveProtocolSuite
from .communicator.interface import tagname
from .shamir import ShamirProtocolSuite, ShamirBasicProtocolSuite


class _CallLogger(hunter.actions.Action):
    def __init__(self):
        self.stack = []

        self.display_whitelist = set([
            "cicada.active.ActiveProtocolSuite.absolute",
            "cicada.active.ActiveProtocolSuite.add",
            "cicada.active.ActiveProtocolSuite.bit_compose",
            "cicada.active.ActiveProtocolSuite.bit_decompose",
            "cicada.active.ActiveProtocolSuite.divide",
            "cicada.active.ActiveProtocolSuite.dot",
            "cicada.active.ActiveProtocolSuite.equal",
            "cicada.active.ActiveProtocolSuite.field_add",
            "cicada.active.ActiveProtocolSuite.field_dot",
            "cicada.active.ActiveProtocolSuite.field_multiply",
            "cicada.active.ActiveProtocolSuite.field_power",
            "cicada.active.ActiveProtocolSuite.field_subtract",
            "cicada.active.ActiveProtocolSuite.field_uniform",
            "cicada.active.ActiveProtocolSuite.floor",
            "cicada.active.ActiveProtocolSuite.less",
            "cicada.active.ActiveProtocolSuite.less_zero",
            "cicada.active.ActiveProtocolSuite.logical_and",
            "cicada.active.ActiveProtocolSuite.logical_not",
            "cicada.active.ActiveProtocolSuite.logical_or",
            "cicada.active.ActiveProtocolSuite.logical_xor",
            "cicada.active.ActiveProtocolSuite.maximum",
            "cicada.active.ActiveProtocolSuite.minimum",
            "cicada.active.ActiveProtocolSuite.multiplicative_inverse",
            "cicada.active.ActiveProtocolSuite.multiply",
            "cicada.active.ActiveProtocolSuite.negative",
            "cicada.active.ActiveProtocolSuite.power",
            "cicada.active.ActiveProtocolSuite.random_bitwise_secret",
            "cicada.active.ActiveProtocolSuite.relu",
            "cicada.active.ActiveProtocolSuite.reshare",
            "cicada.active.ActiveProtocolSuite.reveal",
            "cicada.active.ActiveProtocolSuite.right_shift",
            "cicada.active.ActiveProtocolSuite.share",
            "cicada.active.ActiveProtocolSuite.subtract",
            "cicada.active.ActiveProtocolSuite.sum",
            "cicada.active.ActiveProtocolSuite.zigmoid",

            "cicada.additive.AdditiveProtocolSuite.absolute",
            "cicada.additive.AdditiveProtocolSuite.add",
            "cicada.additive.AdditiveProtocolSuite.bit_compose",
            "cicada.additive.AdditiveProtocolSuite.bit_decompose",
            "cicada.additive.AdditiveProtocolSuite.divide",
            "cicada.additive.AdditiveProtocolSuite.dot",
            "cicada.additive.AdditiveProtocolSuite.equal",
            "cicada.additive.AdditiveProtocolSuite.field_add",
            "cicada.additive.AdditiveProtocolSuite.field_dot",
            "cicada.additive.AdditiveProtocolSuite.field_multiply",
            "cicada.additive.AdditiveProtocolSuite.field_power",
            "cicada.additive.AdditiveProtocolSuite.field_subtract",
            "cicada.additive.AdditiveProtocolSuite.field_uniform",
            "cicada.additive.AdditiveProtocolSuite.floor",
            "cicada.additive.AdditiveProtocolSuite.less",
            "cicada.additive.AdditiveProtocolSuite.less_zero",
            "cicada.additive.AdditiveProtocolSuite.logical_and",
            "cicada.additive.AdditiveProtocolSuite.logical_not",
            "cicada.additive.AdditiveProtocolSuite.logical_or",
            "cicada.additive.AdditiveProtocolSuite.logical_xor",
            "cicada.additive.AdditiveProtocolSuite.maximum",
            "cicada.additive.AdditiveProtocolSuite.minimum",
            "cicada.additive.AdditiveProtocolSuite.multiplicative_inverse",
            "cicada.additive.AdditiveProtocolSuite.multiply",
            "cicada.additive.AdditiveProtocolSuite.negative",
            "cicada.additive.AdditiveProtocolSuite.power",
            "cicada.additive.AdditiveProtocolSuite.random_bitwise_secret",
            "cicada.additive.AdditiveProtocolSuite.relu",
            "cicada.additive.AdditiveProtocolSuite.reshare",
            "cicada.additive.AdditiveProtocolSuite.reveal",
            "cicada.additive.AdditiveProtocolSuite.right_shift",
            "cicada.additive.AdditiveProtocolSuite.share",
            "cicada.additive.AdditiveProtocolSuite.subtract",
            "cicada.additive.AdditiveProtocolSuite.sum",
            "cicada.additive.AdditiveProtocolSuite.zigmoid",

            "cicada.shamir.ShamirBasicProtocolSuite.add",
            "cicada.shamir.ShamirBasicProtocolSuite.bit_compose",
            "cicada.shamir.ShamirBasicProtocolSuite.field_add",
            "cicada.shamir.ShamirBasicProtocolSuite.field_subtract",
            "cicada.shamir.ShamirBasicProtocolSuite.field_uniform",
            "cicada.shamir.ShamirBasicProtocolSuite.negative",
            "cicada.shamir.ShamirBasicProtocolSuite.reshare",
            "cicada.shamir.ShamirBasicProtocolSuite.reveal",
            "cicada.shamir.ShamirBasicProtocolSuite.share",
            "cicada.shamir.ShamirBasicProtocolSuite.subtract",
            "cicada.shamir.ShamirBasicProtocolSuite.sum",

            "cicada.shamir.ShamirProtocolSuite.absolute",
            "cicada.shamir.ShamirProtocolSuite.add",
            "cicada.shamir.ShamirProtocolSuite.bit_compose",
            "cicada.shamir.ShamirProtocolSuite.bit_decompose",
            "cicada.shamir.ShamirProtocolSuite.divide",
            "cicada.shamir.ShamirProtocolSuite.dot",
            "cicada.shamir.ShamirProtocolSuite.equal",
            "cicada.shamir.ShamirProtocolSuite.field_add",
            "cicada.shamir.ShamirProtocolSuite.field_dot",
            "cicada.shamir.ShamirProtocolSuite.field_multiply",
            "cicada.shamir.ShamirProtocolSuite.field_power",
            "cicada.shamir.ShamirProtocolSuite.field_subtract",
            "cicada.shamir.ShamirProtocolSuite.field_uniform",
            "cicada.shamir.ShamirProtocolSuite.floor",
            "cicada.shamir.ShamirProtocolSuite.less",
            "cicada.shamir.ShamirProtocolSuite.less_zero",
            "cicada.shamir.ShamirProtocolSuite.logical_and",
            "cicada.shamir.ShamirProtocolSuite.logical_not",
            "cicada.shamir.ShamirProtocolSuite.logical_or",
            "cicada.shamir.ShamirProtocolSuite.logical_xor",
            "cicada.shamir.ShamirProtocolSuite.maximum",
            "cicada.shamir.ShamirProtocolSuite.minimum",
            "cicada.shamir.ShamirProtocolSuite.multiplicative_inverse",
            "cicada.shamir.ShamirProtocolSuite.multiply",
            "cicada.shamir.ShamirProtocolSuite.negative",
            "cicada.shamir.ShamirProtocolSuite.power",
            "cicada.shamir.ShamirProtocolSuite.random_bitwise_secret",
            "cicada.shamir.ShamirProtocolSuite.relu",
            "cicada.shamir.ShamirProtocolSuite.reshare",
            "cicada.shamir.ShamirProtocolSuite.reveal",
            "cicada.shamir.ShamirProtocolSuite.right_shift",
            "cicada.shamir.ShamirProtocolSuite.share",
            "cicada.shamir.ShamirProtocolSuite.subtract",
            "cicada.shamir.ShamirProtocolSuite.sum",
            "cicada.shamir.ShamirProtocolSuite.zigmoid",
            ])

        self.test_whitelist = set([
            "cicada.arithmetic.Field.__call__",
            "cicada.arithmetic.Field.add",
            "cicada.arithmetic.Field.full_like",
            "cicada.arithmetic.Field.inplace_add",
            "cicada.arithmetic.Field.inplace_subtract",
            "cicada.arithmetic.Field.multiply",
            "cicada.arithmetic.Field.negative",
            "cicada.arithmetic.Field.ones",
            "cicada.arithmetic.Field.ones_like",
            "cicada.arithmetic.Field.subtract",
            "cicada.arithmetic.Field.sum",
            "cicada.arithmetic.Field.uniform",
            "cicada.arithmetic.Field.zeros",
            "cicada.arithmetic.Field.zeros_like",
            "cicada.encoding.Bits.decode",
            "cicada.encoding.Bits.encode",
            "cicada.encoding.Boolean.decode",
            "cicada.encoding.Boolean.encode",
            "cicada.encoding.FixedPoint.decode",
            "cicada.encoding.FixedPoint.encode",
            "cicada.encoding.Identity.decode",
            "cicada.encoding.Identity.encode",
            ])


    def __call__(self, event):
        if not hasattr(event.function_object, "__qualname__"):
            return

        fqname = event.module + "." + event.function_object.__qualname__
        qname = event.function_object.__qualname__
        name = event.function_object.__name__

        #####################################################################################
        # Log sent messages

        if event.kind == "call" and fqname == "cicada.communicator.socket.SocketCommunicator._send":
            communicator = event.locals["self"]
            dst = event.locals["dst"]
            payload = event.locals["payload"]
            tag = event.locals["tag"]

            net = Message()
            net.arrow = "-->"
            net.comm = communicator
            net.dir = ">"
            net.dst = dst
            net.other = dst
            net.payload = payload
            net.src = communicator.rank
            net.tag = tagname(tag)
            net.verb = "sent"

            logger.info("Sent message", extra={"net": net})

        #####################################################################################
        # Log received messages

        if event.kind == "call" and fqname == "cicada.communicator.socket.SocketCommunicator._queue_message":
            communicator = event.locals["self"]
            src = event.locals["src"]
            payload = event.locals["payload"]
            tag = event.locals["tag"]

            net = Message()
            net.arrow = "<--"
            net.comm = communicator
            net.dir = "<"
            net.dst = communicator.rank
            net.other = src
            net.payload = payload
            net.src = src
            net.tag = tagname(tag)
            net.verb = "received"

            logger.info("Received message", extra={"net": net})

        #####################################################################################
        # Log consistency verification code

        # Hide __init__ functions.
        if name in ["__init__"]:
            return

        # Hide private functions.
        if name.startswith("_") and not name.startswith("__"):
            return

        # Identify functions that should be displayed in the transcript.
        display = True if fqname in self.display_whitelist else False

        # Identify functions that should generate consistency verification code in the transcript.
        test = True if fqname in self.test_whitelist else False

        # Convert display function calls into comments.
        if event.kind == "call" and display:
            args = event.locals
            o = self.repr(args["self"])
            signature = ", ".join([f"{key}={self.repr(value)}" for key, value in args.items() if key != "self"])
            _log_code(event, f"# {o}.{name}({signature})", first=True, last=True)
            return

        # Make copies of test function call arguments, in-case they're modified in-place.
        if event.kind == "call" and test:
            args = {}
            for key, value in event.locals.items():
                if key in ["self"]:
                    args[key] = value
                else:
                    args[key] = copy.deepcopy(value)

            self.stack.append(args)
            return

        # Convert test function calls into consistency verification statements.
        if event.kind == "return" and test:
            args = self.stack.pop()
            locals = event.locals
            result = event.arg

            if fqname in ["cicada.arithmetic.Field.inplace_add", "cicada.arithmetic.Field.inplace_subtract"]:
                o = self.repr(args["self"])
                lhs = self.repr(args["lhs"])
                rhs = self.repr(args["rhs"])
                result = self.repr(locals["lhs"])

                _log_code(event, f"lhs = {lhs}", first=True)
                _log_code(event, f"{o}.{name}(lhs=lhs, rhs={rhs})")
                _log_code(event, f"cicada.transcript.assert_equal(lhs, {result})", last=True)

            elif fqname == "cicada.arithmetic.Field.uniform":
                o = self.repr(args["self"])
                size = self.repr(args["size"])
                bg = self.repr(args["generator"].bit_generator)
                state = self.repr(args["generator"].bit_generator.state)
                result = self.repr(result)

                _log_code(event, f"bg = {bg}", first=True)
                _log_code(event, f"bg.state = {state}")
                _log_code(event, f"cicada.transcript.assert_equal({o}.{name}(size={size}, generator=numpy.random.Generator(bg)), {result})", last=True)

            elif "self" in args:
                o = self.repr(args["self"])
                signature = ", ".join([f"{key}={self.repr(value)}" for key, value in args.items() if key != "self"])
                result = self.repr(result)

                _log_code(event, f"cicada.transcript.assert_equal({o}.{name}({signature}), {result})", first=True, last=True)

            else:
                signature = ", ".join([f"{key}={self.repr(value)}" for key, value in args.items()])
                result = self.repr(result)

                _log_code(event, f"cicada.transcript.assert_equal({name}({signature}), {result})", first=True, last=True)


    def repr(self, o):
        if isinstance(o, ActiveProtocolSuite):
            return f"cicada.active.ActiveProtcolSuite()"
        if isinstance(o, AdditiveProtocolSuite):
            return f"cicada.additive.AdditiveProtocolSuite()"
        if isinstance(o, ShamirBasicProtocolSuite):
            return f"cicada.additive.ShamirBasicProtocolSuite()"
        if isinstance(o, ShamirProtocolSuite):
            return f"cicada.additive.ShamirProtocolSuite()"
        if isinstance(o, numpy.ndarray):
            return f"numpy.array({o.tolist()}, dtype='{o.dtype}')"
        if isinstance(o, numpy.random.Generator):
            return f"numpy.random.Generator({self.repr(o.bit_generator)})"
        if isinstance(o, numpy.random.PCG64):
            return f"numpy.random.PCG64()"
        return repr(o)


class Code(types.SimpleNamespace):
    """Stores code-related metadata for use in :class:`Formatter`."""
    pass


class Formatter(object):
    """Custom log formatter for transcription messages.

    Unlike the generic :class:`logging.Formatter` objects provided with Python, this
    class is configured with more than one format string, to handle each type of
    event that can be logged.  In addition to the standard format fields provided
    by :ref:`logrecord-attributes`, the format strings can use the following.

    For sent-message and received-message events:

    * net.arrow - message direction, relative to the player logging the event.
    * net.comm - communicator that sent / received the message.
    * net.comm.name - name of the communicator that sent / received the message.
    * net.comm.rank - rank of the player logging the event.
    * net.dir - message direction, relative to the player logging the event.
    * net.dst - rank of the player receiving the message.
    * net.other - rank of the player sending or receiving with the player logging the event.
    * net.payload - message payload contents.
    * net.src - rank of the player sending the message.
    * net.tag - message type.
    * net.verb - message direction, relative to the player logging the event.

    For code events:

    * code.filename - the full path to the file containing the original statement.
    * code.first - this is the first line of generated code associated with the original statement.
    * code.last - this is the last line of generated code associated with the original statement.
    * code.lineno - line number of the file containing the original statement.

    Parameters
    ----------
    fmt: :class:`str`, optional
        Format string for context records.
    netfmt: :class:`str`, optional
        Format string for sent- and received-message records.
    codefmt: :class:`str`, optional
        Format string for consistency verification records.
    """
    def __init__(self, fmt, netfmt, codefmt, codepre, codepost):
        self._fmt = fmt
        self._netfmt = netfmt
        self._codefmt = codefmt
        self._codepre = codepre
        self._codepost = codepost


    def format(self, record):
        """Formats a log record for display."""

        # Format network messages.
        if hasattr(record, "net"):
            msg = self._netfmt.format_map(record.__dict__)
            return msg

        # Format code.
        if hasattr(record, "code"):
            msg = self._codefmt.format_map(record.__dict__)
            if self._codepre and record.code.first:
                msg = self._codepre.format_map(record.__dict__) + msg
            if self._codepost and record.code.last:
                msg = msg + self._codepost.format_map(record.__dict__)
            return msg

        # Format generic messages.
        msg = self._fmt.format_map(record.__dict__)
        return msg


class HideCode(object):
    """Log filter that hides code records."""
    def filter(self, record):
        if hasattr(record, "code"):
            return False
        return True


class HideContextMessages(object):
    """Log filter that hides context message records."""
    def filter(self, record):
        if not hasattr(record, "code") and not hasattr(record, "net"):
            return False
        return True


class HideReceivedMessages(object):
    """Log filter that hides network message records for received messages."""
    def filter(self, record):
        if hasattr(record, "net") and record.net.verb == "received":
            return False
        return True


class HideSentMessages(object):
    """Log filter that hides network message records for sent messages."""
    def filter(self, record):
        if hasattr(record, "net") and record.net.verb == "sent":
            return False
        return True


class Message(types.SimpleNamespace):
    """Stores message-related metadata for use in :class:`Formatter`."""
    pass


def _log_code(event, message, first=False, last=False):
    code = Code()
    code.filename = event.filename
    code.lineno = event.lineno
    code.first = first
    code.last = last

    logger.info(message, extra={"code": code})


def assert_equal(lhs, rhs):
    """Test two objects for equality.

    Seamlessly handles special types such as numpy.ndarray, etc.
    """
    if isinstance(lhs, numpy.ndarray) or isinstance(rhs, numpy.ndarray):
        if not numpy.array_equal(lhs, rhs):
            raise AssertionError(f"{lhs} != {rhs}")
        return

    if not lhs == rhs:
        raise AssertionError(f"{lhs} != {rhs}")


def code_handler(handler=None, fmt=None, netfmt=None, codefmt=None, codepre=None, codepost=None, sent=False, received=False):
    """Create a :class:`logging.Handler`, configured to display consistency verification code records.

    Parameters
    ----------
    handler: :class:`logging.Handler`, optional
        The handler to be configured. Defaults to a new instance of
        :class:`logging.StreamHandler` if :any:`None`.
    fmt: :class:`str`, optional
        Format string for context records.
    netfmt: :class:`str`, optional
        Format string for sent- and received-message records.
    codefmt: :class:`str`, optional
        Format string for consistency verification records.
    codepre: :class:`str`, optional
        Format string displayed before each group of consistency verification code records.
    codepost: :class:`str`, optional
        Format string displayed after each group of consistency verification code records.
    sent: :any:`bool`, optional
        Enable displaying sent-message records if :any:`True`.
    received: :any:`bool`, optional
        Enable displaying received-message records if :any:`True`.
    """
    if handler is None:
        handler = logging.StreamHandler()

    if fmt is None:
        fmt = "# {msg}"
    if netfmt is None:
        netfmt = "# {net.comm.rank} {net.arrow} {net.other} {net.tag} {net.payload}"
    if codefmt is None:
        codefmt = "{msg}"
    if codepost is None:
        codepost = "\n"

    if not sent:
        handler.addFilter(HideSentMessages())
    if not received:
        handler.addFilter(HideReceivedMessages())
    handler.setFormatter(Formatter(fmt=fmt, netfmt=netfmt, codefmt=codefmt, codepre=codepre, codepost=codepost))

    return handler


def net_handler(handler=None, fmt=None, netfmt=None, codefmt=None, codepre=None, codepost=None, sent=True, received=True, code=False):
    """Create a :class:`logging.Handler`, configured to display network message records.

    Parameters
    ----------
    handler: :class:`logging.Handler`, optional
        The handler to be configured. Defaults to a new instance of
        :class:`logging.StreamHandler` if :any:`None`.
    fmt: :class:`str`, optional
        Format string for context records.
    netfmt: :class:`str`, optional
        Format string for sent- and received-message records.
    codefmt: :class:`str`, optional
        Format string for consistency verification records.
    codepre: :class:`str`, optional
        Format string displayed before each group of consistency verification code records.
    codepost: :class:`str`, optional
        Format string displayed after each group of consistency verification code records.
    sent: :any:`bool`, optional
        Disable displaying sent-message records if :any:`False`.
    received: :any:`bool`, optional
        Disable displaying received-message records if :any:`False`.
    code: :any:`bool`, optional
        Display consistency verification code records if :any:`True`.
    """
    if handler is None:
        handler = logging.StreamHandler()

    if fmt is None:
        fmt = "{processName}: {msg}"
    if netfmt is None:
        netfmt = "{processName}: {net.arrow} {net.other} {net.tag} {net.payload}"
    if codefmt is None:
        codefmt = "{processName}: {msg}"

    if not code:
        handler.addFilter(HideCode())
    if not sent:
        handler.addFilter(HideSentMessages())
    if not received:
        handler.addFilter(HideReceivedMessages())
    handler.setFormatter(Formatter(fmt=fmt, netfmt=netfmt, codefmt=codefmt, codepre=codepre, codepost=codepost))

    return handler


def set_handler(logger, handler):
    """Set the handler for a logger, removing any other handlers.

    Parameters
    ----------
    logger: :class:`logging.Logger`, required
        The logger to be modified.
    handler: :class:`logging.Handler`, required
        The handler to be assigned to `logger`.
    """
    logger.level = logging.INFO
    while logger.handlers:
        logger.removeHandler(logger.handlers[0])
    logger.addHandler(handler)


def log(message=None):
    """Log general-purpose events into the transcription.

    Application code should use this to incorporate high-level application
    context information alongside the lower-level network and consistency
    verification transcript contents.
    """
    logger.info(message)


def record():
    """Enable transcription.

    All transcription functionality depends on tracing function calls, so this must
    be called to begin transcription.  The result is a context manager that can be
    used in with-statements.
    """
    return hunter.trace(module_startswith="cicada", kind_in=("call", "return"), action=_CallLogger())


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)




# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import test

from behave import *
import numpy

import cicada.arithmetic


@given(u'a default Field')
def step_impl(context):
    if "fields" not in context:
        context.fields = []
    context.fields.append(cicada.arithmetic.Field())


@given(u'a Field with order {order}')
def step_impl(context, order):
    order = eval(order)
    if "fields" not in context:
        context.fields = []
    context.fields.append(cicada.arithmetic.Field(order=order))


@given(u'a field array {x}')
def step_impl(context, x):
    x = numpy.array(eval(x))
    field = context.fields[-1]
    if "fieldarrays" not in context:
        context.fieldarrays = []
    context.fieldarrays.append(field(x))


@when(u'a field with order {order} is created, it should not raise an exception')
def step_impl(context, order):
    order = eval(order)
    field = cicada.arithmetic.Field(order=order)


@when(u'a field with order {order} is created, it should raise an exception')
def step_impl(context, order):
    order = eval(order)
    with unittest.TestCase().assertRaises(ValueError) as cm:
        field = cicada.arithmetic.Field(order=order)
    print(cm.exception)


@when(u'generating a field array of uniform random values with shape {shape}')
def step_impl(context, shape):
    shape = eval(shape)
    field = context.fields[-1]
    generator = numpy.random.default_rng()
    if "fieldarrays" not in context:
        context.fieldarrays = []
    context.fieldarrays.append(field.uniform(size=shape, generator=generator))


@when(u'generating a field array of zeros with shape {shape}')
def step_impl(context, shape):
    shape = eval(shape)
    field = context.fields[-1]
    if "fieldarrays" not in context:
        context.fieldarrays = []
    context.fieldarrays.append(field.zeros(shape))


@when(u'generating a field array of zeros like {other}')
def step_impl(context, other):
    other = numpy.array(eval(other))
    field = context.fields[-1]
    if "fieldarrays" not in context:
        context.fieldarrays = []
    context.fieldarrays.append(field.zeros_like(other))


@when(u'the field array is negated')
def step_impl(context):
    field = context.fields[-1]
    fieldarray = context.fieldarrays.pop()
    context.fieldarrays.append(field.negative(fieldarray))


@when(u'the field array is summed')
def step_impl(context):
    field = context.fields[-1]
    fieldarray = context.fieldarrays.pop()
    context.fieldarrays.append(field.sum(fieldarray))


@when(u'the second field array is subtracted from the first')
def step_impl(context):
    field = context.fields[-1]
    b = context.fieldarrays.pop()
    a = context.fieldarrays.pop()
    context.fieldarrays.append(field.subtract(a, b))


@when(u'the second field array is added to the first')
def step_impl(context):
    field = context.fields[-1]
    b = context.fieldarrays.pop()
    a = context.fieldarrays.pop()
    context.fieldarrays.append(field.add(a, b))


@when(u'the second field array is added in-place to the first')
def step_impl(context):
    field = context.fields[-1]
    b = context.fieldarrays.pop()
    a = context.fieldarrays[-1]
    field.inplace_add(a, b)


@when(u'the second field array is subtracted in-place from the first')
def step_impl(context):
    field = context.fields[-1]
    b = context.fieldarrays.pop()
    a = context.fieldarrays[-1]
    field.inplace_subtract(a, b)


@then(u'the field array should match {result}')
def step_impl(context, result):
    field = context.fields[-1]
    result = field(eval(result))
    fieldarray = context.fieldarrays[-1]
    numpy.testing.assert_array_equal(fieldarray, result)


@then(u'the fields should compare equal')
def step_impl(context):
    lhs, rhs = context.fields
    test.assert_true(lhs == rhs)


@then(u'the fields should compare unequal')
def step_impl(context):
    lhs, rhs = context.fields
    test.assert_true(lhs != rhs)


@then(u'the field array shape should match {shape}')
def step_impl(context, shape):
    shape = eval(shape)
    fieldarray = context.fieldarrays[-1]
    test.assert_equal(shape, fieldarray.shape)


@then(u'the field array values should be in-range for the field')
def step_impl(context):
    field = context.fields[-1]
    fieldarray = context.fieldarrays[-1]
    test.assert_true(numpy.all(fieldarray >= 0))
    test.assert_true(numpy.all(fieldarray < field.order))






# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality to make logging from multiple processes easier."""

import contextlib
import logging
import numbers

from cicada.communicator.interface import Communicator, Tag


class Logger(object):
    """Wrap a normal Python logger with a Cicada communicator to synchronize player output.

    Note
    ----
    Because :class:`Logger` communicates among players for synchronization, it
    can have a significant effect on performance, even when log messages are
    discarded by log levels or other filtering.

    Furthermore, :class:`Logger` should not be used in error recovery code, since it will
    fail attempting to communicate with players that are (presumably) dead.

    You can pass `sync=False` when creating :class:`Logger` to disable synchronization,
    e.g. if you're running your code on separate hosts or in separate terminal sessions.

    Parameters
    ----------
    logger: :class:`logging.Logger`, required.
        The Python logger to be used for output.
    communicator: :class:`cicada.communicator.interface.Communicator`, required
        The communicator that will be used to synchronize output among players.
    sync: :class:`bool`, optional
        Used to control synchronization, which is enabled by default.
    """
    def __init__(self, logger, communicator, sync=True):
        if not isinstance(communicator, Communicator):
            raise ValueError("A Cicada communicator is required.") # pragma: no cover

        self._logger = logger
        self._communicator = communicator
        self._sync = sync


    @property
    def communicator(self):
        """Returns the underlying communicator."""
        return self._communicator


    def critical(self, msg, *args, src=None, **kwargs):
        """Log a critical message, synchronized among players.

        .. note::

             This is a collective operation that *must* be called by all players that are members of the communicator.

        The arguments match those of :meth:`logging.Logger.critical`, with the addition of the following:

        Parameters
        ----------
        src: :class:`int` or sequence of :class:`int`, optional
            If specified, only the given player(s) will produce log output.
        """
        self.log(logging.CRITICAL, msg, *args, src=src, **kwargs)


    def debug(self, msg, *args, src=None, **kwargs):
        """Log a debug message, synchronized among players.

        .. note::

             This is a collective operation that *must* be called by all players that are members of the communicator.

        The arguments match those of :meth:`logging.Logger.debug`, with the addition of the following:

        Parameters
        ----------
        src: :class:`int` or sequence of :class:`int`, optional
            If specified, only the given player(s) will produce log output.
        """
        self.log(logging.DEBUG, msg, *args, src=src, **kwargs)


    def error(self, msg, *args, src=None, **kwargs):
        """Log an error message, synchronized among players.

        .. note::

             This is a collective operation that *must* be called by all players that are members of the communicator.

        The arguments match those of :meth:`logging.Logger.error`, with the addition of the following:

        Parameters
        ----------
        src: :class:`int` or sequence of :class:`int`, optional
            If specified, only the given player(s) will produce log output.
        """
        self.log(logging.ERROR, msg, *args, src=src, **kwargs)


    def info(self, msg, *args, src=None, **kwargs):
        """Log an info message, synchronized among players.

        .. note::

             This is a collective operation that *must* be called by all players that are members of the communicator.

        The arguments match those of :meth:`logging.Logger.info`, with the addition of the following:

        Parameters
        ----------
        src: :class:`int` or sequence of :class:`int`, optional
            If specified, only the given player(s) will produce log output.
        """
        self.log(logging.INFO, msg, *args, src=src, **kwargs)


    def log(self, level, msg, *args, src=None, **kwargs):
        """Log a message, synchronized among players.

        .. note::

             This is a collective operation that *must* be called by all players that are members of the communicator.

        The arguments match those of :meth:`logging.Logger.log`, with the addition of the following:

        Parameters
        ----------
        src: :class:`int` or sequence of :class:`int`, optional
            If specified, only the given player(s) will produce log output.
        """
        communicator = self._communicator

        if src is None:
            src = communicator.ranks
        elif isinstance(src, numbers.Integral):
            src = [src]

        # Wait for our turn to generate output.
        if self._sync and communicator.rank:
            communicator.recv(src=communicator.rank-1, tag=Tag.LOGSYNC)

        # Generate output.
        if communicator.rank in src:
            self._logger.log(level, msg, *args, **kwargs)

        # Notify the next player that it's their turn.
        if self._sync and communicator.rank < communicator.world_size-1:
            communicator.send(dst=communicator.rank+1, value=None, tag=Tag.LOGSYNC)

        # The last player notifies the group that the output is complete.
        if self._sync and communicator.rank == communicator.world_size-1:
            for rank in communicator.ranks:
                communicator.send(dst=rank, value=None, tag=Tag.LOGSYNC)

        # Wait until output is complete before we return.
        if self._sync:
            communicator.recv(src=communicator.world_size-1, tag=Tag.LOGSYNC)


    @property
    def logger(self):
        """Returns the underlying Python :class:`logging.Logger`."""
        return self._logger


    @contextlib.contextmanager
    def override(self, *, sync=None):
        """Temporarily change logging behavior.

        Use :meth:`override` to temporarily modify logger behavior in a with statement::

            with log.override(sync=False):
                # Do uncoordinated logging here.
            # Go back to coordinated logging here.

        .. note::

            Changes to logging behavior *must* be consistent for *all* players that are members of the communicator.

        Parameters
        ----------
        sync: :class:`bool`, optional
            If specified, override the logger sync property.

        Returns
        -------
        context: :class:`object`
            A context manager object that will restore the loger state when exited.
        """
        original_context = {
            "sync": self._sync,
        }

        try:
            if sync is not None:
                self._sync = sync
            yield original_context
        finally:
            if sync is not None:
                self._sync = original_context["sync"]

    @property
    def sync(self):
        """Controls whether coordinated logging is enabled or not.

        .. note::

            Changes to `sync` *must* be consistent for *all* players that are members of the communicator.
        """
        return self._sync


    @sync.setter
    def sync(self, value):
        self._sync = bool(value)


    def warning(self, msg, *args, src=None, **kwargs):
        """Log a warning message, synchronized among players.

        .. note::

             This is a collective operation that *must* be called by all players that are members of the communicator.

        The arguments match those of :meth:`logging.Logger.warning`, with the addition of the following:

        Parameters
        ----------
        src: :class:`int` or sequence of :class:`int`, optional
            If specified, only the given player(s) will produce log output.
        """
        self.log(logging.WARNING, msg, *args, src=src, **kwargs)




# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy

from cicada.additive import AdditiveProtocolSuite
from cicada.communicator import SocketCommunicator

def main(communicator):
    protocol = AdditiveProtocolSuite(communicator)

    ashare = protocol.share(src=0, secret=numpy.array(2), shape=())
    bshare = protocol.share(src=0, secret=numpy.array(3), shape=())
    print("private-private addition:", protocol.reveal(protocol.add(ashare, bshare)))
    print("public-private addition:", protocol.reveal(protocol.add(numpy.array(1.5), ashare)))
    print("private-public addition:", protocol.reveal(protocol.add(bshare, numpy.array(3.5))))

SocketCommunicator.run(world_size=3, fn=main)
